import argparse, os, sys
from os import listdir
from os.path import isfile, join
from ipywidgets import widgets

import pickle
import math
import collections
import time

import numpy as np
import pandas as pd
import scanpy as sc
import warnings
warnings.filterwarnings('ignore')
from sklearn.metrics import silhouette_score
import multiprocess as mp
from functools import partial
import seaborn as sns
from matplotlib import pyplot as plt
from .._settings import add_reference


import time
def autoResolution(adata,cpus=4):
    r"""
    Automatically determine clustering resolution

    Arguments:
        adata : `scanpy.AnnData` The single cell data.
        cpus : `int`, optional (default: 4) The number of cpus used for parallel computing.
    
    Returns:
        adata : `scanpy.AnnData` The single cell data with the clustering resolution.
        res : `float` The clustering resolution.
        df_sil: `pandas.DataFrame` The silhouette score of each clustering resolution.
    """
    print("Automatically determine clustering resolution...")
    start = time.time()
    def subsample_clustering(adata, sample_n, subsample_n, resolution, subsample):
        subadata = adata[subsample]
        sc.tl.louvain(subadata, resolution=resolution)
        cluster = subadata.obs['louvain'].tolist()
        
        subsampling_n = np.zeros((sample_n, sample_n), dtype=bool)
        coclustering_n = np.zeros((sample_n, sample_n), dtype=bool)
        
        for i in range(subsample_n):
            for j in range(subsample_n):
                x = subsample[i]
                y = subsample[j]
                subsampling_n[x][y] = True
                if cluster[i] == cluster[j]:
                    coclustering_n[x][y] = True
        return (subsampling_n, coclustering_n)
    rep_n = 5
    subset = 0.8
    sample_n = len(adata.obs)
    subsample_n = int(sample_n * subset)
    resolutions = np.linspace(0.4, 1.4, 6)
    silhouette_avg = {}
    np.random.seed(1)
    best_resolution = 0
    highest_sil = 0
    for r in resolutions:
        r = np.round(r, 1)
        print("Clustering test: resolution = ", r)
        sub_start = time.time()
        subsamples = [np.random.choice(sample_n, subsample_n, replace=False) for t in range(rep_n)]
        p = mp.Pool(cpus)
        func = partial(subsample_clustering, adata, sample_n, subsample_n, r)
        resultList = p.map(func, subsamples)
        p.close()
        p.join()
        
        subsampling_n = sum([result[0] for result in resultList])
        coclustering_n = sum([result[1] for result in resultList])
        
        subsampling_n[np.where(subsampling_n == 0)] = 1e6
        distance = 1.0 - coclustering_n / subsampling_n
        np.fill_diagonal(distance, 0.0)
        
        sc.tl.louvain(adata, resolution=r, key_added = 'louvain_r' + str(r))
        silhouette_avg[str(r)] = silhouette_score(distance, adata.obs['louvain_r' + str(r)], metric="precomputed")
        if silhouette_avg[str(r)] > highest_sil:
            highest_sil = silhouette_avg[str(r)]
            best_resolution = r
        print("robustness score = ", silhouette_avg[str(r)])
        sub_end = time.time()
        print('time: {}', sub_end - sub_start)
        print()
    adata.obs['louvain'] = adata.obs['louvain_r' + str(best_resolution)]
    print("resolution with highest score: ", best_resolution)
    res = best_resolution
    # write silhouette record to uns and remove the clustering results except for the one with the best resolution
    adata.uns['sihouette score'] = silhouette_avg
    # draw lineplot
    df_sil = pd.DataFrame(silhouette_avg.values(), columns=['silhouette score'], index=[float(x) for x in silhouette_avg.keys()])
    df_sil.plot.line(style='.-', color='green', title='Auto Resolution', xticks=resolutions, xlabel='resolution', ylabel='silhouette score', legend=False)
    #pp.savefig()
    #plt.close()
    end = time.time()
    print('time: {}', end-start)
    return adata, res, df_sil

def writeGEP(adata_GEP,path):
    r"""
    Write the gene expression profile to a file

    Arguments:
        adata_GEP : `scanpy.AnnData` The single cell data with gene expression profile. 
        path : `str` The path to save the gene expression profile.
    
    Returns:

    """
    print('Exporting GEP...')
    sc.pp.normalize_total(adata_GEP, target_sum=1e6)
    mat = adata_GEP.X.transpose()
    if type(mat) is not np.ndarray:
        mat = mat.toarray()
    GEP_df = pd.DataFrame(mat, index=adata_GEP.var.index)
    GEP_df.columns = adata_GEP.obs['louvain'].tolist()
    # GEP_df = GEP_df.loc[adata.var.index[adata.var.highly_variable==True]]
    GEP_df.dropna(axis=1, inplace=True)
    GEP_df.to_csv(os.path.join(path, 'GEP.txt'), sep='\t')
    
class Drug_Response:
    r"""
    Drug_Response class for drug response prediction.
    The raw code could be found at https://github.com/ailabstw/scDrug
    """
    def __init__(self,adata,scriptpath,modelpath,output='./',model='GDSC',clusters='All',
                 cell='A549',cpus=4,n_drugs=10):
        
        """
        Initializes the Drug_Response class.

        Arguments:
            adata: Annotated data matrix with cells as rows and genes as columns.
            scriptpath: Path to the directory containing the CaDRReS scripts for the analysis. You need to download the scripts according to `git clone https://github.com/CSB5/CaDRReS-Sc.git` and set the path to the directory.
            modelpath: Path to the directory containing the pre-trained models.
                You need to download the models according to `Pyomic.utils.download_GDSC_data()` and `Pyomic.utils.download_CaDRReS_model()` and set the path to the directory.
            output: Path to the directory where the output files will be saved (default: './').
            model: The name of the pre-trained model to be used for the analysis (default: 'GDSC').
            clusters: The cluster labels to be used for the analysis. Default is all cells (default: 'All').
            cell: The cell line to be analyzed (default: 'A549').
            cpus: The number of CPUs to be used for the analysis (default: 4).
            n_drugs: The number of top drugs to be selected based on the predicted sensitivity (default: 10).

        Returns:
            None
        """

        self.model = model
        self.adata=adata
        self.clusters=clusters
        self.output=output
        self.n_drugs=n_drugs
        self.modelpath=modelpath

        self.scriptpath = scriptpath
        sys.path.append(os.path.abspath(scriptpath))

        from cadrres_sc import pp, model, evaluation, utility
        
        self.load_model()
        self.drug_info()
        self.bulk_exp()
        self.sc_exp()
        self.kernel_feature_preparartion()
        self.sensitivity_prediction()
        if self.model == 'GDSC':
            self.masked_drugs = list(pd.read_csv(self.modelpath+'masked_drugs.csv')['GDSC'].dropna().astype('int64').astype('str'))
            self.cell_death_proportion()
        else:
            self.masked_drugs = list(pd.read_csv(self.modelpath+'masked_drugs.csv')['PRISM'])
        self.output_result()
        self.figure_output()

    def load_model(self):
        r"""
        load the pre-trained model.

        """
        from cadrres_sc import pp, model, evaluation, utility
        ### IC50/AUC prediction
        ## Read pre-trained model
        #model_dir = '/Users/fernandozeng/Desktop/analysis/scDrug/CaDRReS-Sc-model/'
        model_dir = self.modelpath
        obj_function = widgets.Dropdown(options=['cadrres-wo-sample-bias', 'cadrres-wo-sample-bias-weight'], description='Objetice function')
        self.model_spec_name = obj_function.value
        if self.model == 'GDSC':
            model_file = model_dir + '{}_param_dict_all_genes.pickle'.format(self.model_spec_name)
        elif self.model == 'PRISM':
            model_file = model_dir + '{}_param_dict_prism.pickle'.format(self.model_spec_name)
        else:
            sys.exit('Wrong model name.')
        self.cadrres_model = model.load_model(model_file)

    def drug_info(self):
        r"""
        read the drug information.

        """
        ## Read drug information
        if self.model == 'GDSC':
            self.drug_info_df = pd.read_csv(self.scriptpath + '/preprocessed_data/GDSC/drug_stat.csv', index_col=0)
            self.drug_info_df.index = self.drug_info_df.index.astype(str)
        else:
            self.drug_info_df = pd.read_csv(self.scriptpath + '/preprocessed_data/PRISM/PRISM_drug_info.csv', index_col='broad_id')
        
    def bulk_exp(self):
        r"""
        extract the bulk gene expression data.

        """
        ## Read test data
        if self.model == 'GDSC':
            #GDSC_exp exists in the data folder
            files=os.listdir(self.scriptpath + '/data/GDSC')
            if 'GDSC_exp.tsv' not in files:
                self.gene_exp_df = pd.read_csv(self.modelpath + 'GDSC_exp.tsv.gz', sep='\t', index_col=0)
                self.gene_exp_df = self.gene_exp_df.groupby(self.gene_exp_df.index).mean()
            else:
                self.gene_exp_df = pd.read_csv(self.scriptpath + '/data/GDSC/GDSC_exp.tsv', sep='\t', index_col=0)
                self.gene_exp_df = self.gene_exp_df.groupby(self.gene_exp_df.index).mean()
        else:
            self.gene_exp_df = pd.read_csv(self.scriptpath + '/data/CCLE/CCLE_expression.csv', low_memory=False, index_col=0).T
            self.gene_exp_df.index = [gene.split(sep=' (')[0] for gene in self.gene_exp_df.index]

    def sc_exp(self):
        r"""
        Load cluster-specific gene expression profile
        """
        ## Load cluster-specific gene expression profile
        if self.clusters == 'All':
            clusters = sorted(self.adata.obs['louvain'].unique(), key=int)
        else:
            clusters = [x.strip() for x in self.clusters.split(',')]

        self.cluster_norm_exp_df = pd.DataFrame(columns=clusters, index=self.adata.raw.var.index)
        for cluster in clusters:
            self.cluster_norm_exp_df[cluster] =  self.adata.raw.X[self.adata.obs['louvain']==cluster].mean(axis=0).T \
                                                 if np.sum(self.adata.raw.X[self.adata.obs['louvain']==cluster]) else 0.0

    def kernel_feature_preparartion(self):
        r"""
        kernel feature preparation

        """
        from cadrres_sc import pp, model, evaluation, utility
        ## Read essential genes list
        if self.model == 'GDSC':
            ess_gene_list = self.gene_exp_df.index.dropna().tolist()
        else:
            ess_gene_list = utility.get_gene_list(self.scriptpath + '/preprocessed_data/PRISM/feature_genes.txt')

        ## Calculate fold-change
        cell_line_log2_mean_fc_exp_df, cell_line_mean_exp_df = pp.gexp.normalize_log2_mean_fc(self.gene_exp_df)
            
        self.adata_exp_mean = pd.Series(self.adata.raw.X.mean(axis=0).tolist()[0], index=self.adata.raw.var.index)
        cluster_norm_exp_df = self.cluster_norm_exp_df.sub(self.adata_exp_mean, axis=0)

        ## Calculate kernel feature
        self.test_kernel_df = pp.gexp.calculate_kernel_feature(cluster_norm_exp_df, cell_line_log2_mean_fc_exp_df, ess_gene_list)
    
    def sensitivity_prediction(self):
        r"""
        Predict drug sensitivity
        
        """
        from cadrres_sc import pp, model, evaluation, utility
        ## Drug response prediction
        if self.model == 'GDSC':
            print('...Predicting drug response for using CaDRReS(GDSC): {}'.format(self.model_spec_name))
            self.pred_ic50_df, P_test_df= model.predict_from_model(self.cadrres_model, self.test_kernel_df, self.model_spec_name)
            print('...done!')
        else:
            print('...Predicting drug response for using CaDRReS(PRISM): {}'.format(self.model_spec_name))
            self.pred_auc_df, P_test_df= model.predict_from_model(self.cadrres_model, self.test_kernel_df, self.model_spec_name)
            print('...done!')
        add_reference(self.adata,'scDrug','drug response prediction with CaDRReS')

    def cell_death_proportion(self):
        r"""
        Predict cell death proportion and cell death percentage at the ref_type dosage

        """
        ### Drug kill prediction
        ref_type = 'log2_median_ic50'
        self.drug_list = [x for x in self.pred_ic50_df.columns if not x in self.masked_drugs]
        self.drug_info_df = self.drug_info_df.loc[self.drug_list]
        self.pred_ic50_df = self.pred_ic50_df.loc[:,self.drug_list]

        ## Predict cell death percentage at the ref_type dosage
        pred_delta_df = pd.DataFrame(self.pred_ic50_df.values - self.drug_info_df[ref_type].values, columns=self.pred_ic50_df.columns)
        pred_cv_df = 100 / (1 + (np.power(2, -pred_delta_df)))
        self.pred_kill_df = 100 - pred_cv_df
    
    def output_result(self):
        if self.model == 'GDSC':
            drug_df = pd.DataFrame({'Drug ID': self.drug_list, 
                                    'Drug Name': [self.drug_info_df.loc[drug_id]['Drug Name'] for drug_id in self.drug_list]})
            self.pred_ic50_df = (self.pred_ic50_df.T-self.pred_ic50_df.min(axis=1))/(self.pred_ic50_df.max(axis=1)-self.pred_ic50_df.min(axis=1))
            self.pred_ic50_df = self.pred_ic50_df.T
            self.pred_ic50_df.columns = pd.MultiIndex.from_frame(drug_df)
            self.pred_ic50_df.round(3).to_csv(os.path.join(self.output, 'IC50_prediction.csv'))
            self.pred_kill_df.columns = pd.MultiIndex.from_frame(drug_df)
            self.pred_kill_df.round(3).to_csv(os.path.join(self.output, 'drug_kill_prediction.csv'))
        else:
            drug_list = list(self.pred_auc_df.columns)
            drug_list  = [d for d in drug_list if d not in self.masked_drugs]
            drug_df = pd.DataFrame({'Drug ID':drug_list,
                                    'Drug Name':[self.drug_info_df.loc[d, 'name'] for d in drug_list]})
            self.pred_auc_df = self.pred_auc_df.loc[:,drug_list].T
            self.pred_auc_df = (self.pred_auc_df-self.pred_auc_df.min())/(self.pred_auc_df.max()-self.pred_auc_df.min())
            self.pred_auc_df = self.pred_auc_df.T
            self.pred_auc_df.columns = pd.MultiIndex.from_frame(drug_df)
            self.pred_auc_df.round(3).to_csv(os.path.join(self.output, 'PRISM_prediction.csv'))
    
    def draw_plot(self, df, n_drug=10, name='', figsize=()):
        r"""
        plot heatmap of drug response prediction

        Arguments:
            df : `pandas.DataFrame` drug response prediction dataframe
            n_drug : `int` number of drugs to be plotted
            name : `str` name of the plot
            figsize : `tuple` size of the plot
        """
        def select_drug(df, n_drug):
            selected_drugs = []
            df_tmp = df.reset_index().set_index('Drug Name').iloc[:, 1:]
            for cluster in sorted([x for x in df_tmp.columns], key=int):
                for drug_name in df_tmp.sort_values(by=cluster, ascending=False).index[:n_drug].values:
                    if drug_name not in selected_drugs:
                        selected_drugs.append(drug_name)
            df_tmp = df_tmp.loc[selected_drugs, :]
            return df_tmp

        if self.model == 'GDSC':
            fig, ax = plt.subplots(figsize=figsize) 
            sns.heatmap(df.iloc[:n_drug,:-1], cmap='Blues', \
                        linewidths=0.5, linecolor='lightgrey', cbar=True, cbar_kws={'shrink': .2, 'label': 'Drug Sensitivity'}, ax=ax)
            ax.set_xlabel('Cluster', fontsize=20)
            ax.set_ylabel('Drug', fontsize=20)
            ax.figure.axes[-1].yaxis.label.set_size(20)
            for _, spine in ax.spines.items():
                spine.set_visible(True)
                spine.set_color('lightgrey') 
            plt.savefig(os.path.join(self.output, '{}.png'.format(name)), bbox_inches='tight', dpi=200)
            plt.close()

        else:
            fig, ax = plt.subplots(figsize=(df.shape[1], int(n_drug*df.shape[1]/5))) 
            sns.heatmap(select_drug(df, n_drug), cmap='Reds', \
                        linewidths=0.5, linecolor='lightgrey', cbar=True, cbar_kws={'shrink': .2, 'label': 'Drug Sensitivity'}, ax=ax, vmin=0, vmax=1)
            ax.set_xlabel('Cluster', fontsize=20)
            ax.set_ylabel('Drug', fontsize=20)
            ax.figure.axes[-1].yaxis.label.set_size(20)
            for _, spine in ax.spines.items():
                spine.set_visible(True)
                spine.set_color('lightgrey') 
            plt.savefig(os.path.join(self.output, '{}.png'.format(name)), bbox_inches='tight', dpi=200)
            plt.close()

    def figure_output(self):
        r"""
        plot figures

        """
        print('...Ploting figures...')
        ## GDSC figures
        if self.model == 'GDSC':
            tmp_pred_ic50_df = self.pred_ic50_df.T
            tmp_pred_ic50_df = tmp_pred_ic50_df.assign(sum=tmp_pred_ic50_df.sum(axis=1)).sort_values(by='sum', ascending=True)
            self.draw_plot(tmp_pred_ic50_df, name='GDSC prediction', figsize=(12,40))
            tmp_pred_kill_df = self.pred_kill_df.T
            tmp_pred_kill_df = tmp_pred_kill_df.loc[(tmp_pred_kill_df>=50).all(axis=1)]
            tmp_pred_kill_df = tmp_pred_kill_df.assign(sum=tmp_pred_kill_df.sum(axis=1)).sort_values(by='sum', ascending=False)
            self.draw_plot(tmp_pred_kill_df, n_drug=10, name='predicted cell death', figsize=(12,8))

        ## PRISM figures
        else:
            tmp_pred_auc_df = self.pred_auc_df.T
            #tmp_pred_auc_df = tmp_pred_auc_df.assign(sum=tmp_pred_auc_df.sum(axis=1)).sort_values(by='sum', ascending=True)
            self.draw_plot(tmp_pred_auc_df, n_drug=self.n_drugs, name='PRISM prediction')  
        print('done!')  



import scanpy as sc
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import anndata
import seaborn as sns
from .._settings import add_reference
from ._aucell import derive_auc_threshold,fast_rank,_rank_sparse_row,aucell

#from ctxcore.recovery import enrichment4cells,aucs
#from ctxcore.genesig import GeneSignature

ctxcore_install=False

def check_ctxcore():
    """
    
    """
    global ctxcore_install
    try:
        import ctxcore
        ctxcore_install=True
        print('ctxcore have been install version:',ctxcore.__version__)
    except ImportError:
        raise ImportError(
            'Please install the ctxcore: `pip install ctxcore`.'
        )


def geneset_aucell_tmp(adata, geneset_name, geneset, AUC_threshold=0.01, seed=42, chunk_size=10000):
    """
    Calculate the AUC-ell score for a given gene set.

    Arguments:
        adata: AnnData object
            Annotated data matrix containing gene expression data.
        geneset_name: str
            Name of the gene set.
        geneset: list of str
            List of gene symbols for the gene set.
        AUC_threshold: float, optional
            AUC threshold used to determine significant interactions (default is 0.01).
        seed: int, optional
            Seed used to initialize the random number generator (default is 42).
        chunk_size: int, optional
            The number of cells to process in each chunk (default is 10000).

    Returns:
        None
           Adds a column to the 'obs' attribute of the adata object containing the AUC-ell score for the gene set.
    """

    check_ctxcore()
    global ctxcore_install
    if ctxcore_install == True:
        global aucs
        global GeneSignature
        from ctxcore.recovery import aucs
        from ctxcore.genesig import GeneSignature

    matrix = adata.to_df()
    percentiles = derive_auc_threshold(matrix)
    auc_threshold = percentiles[AUC_threshold]

    n_cells = matrix.shape[0]
    auc_results = np.zeros(n_cells, dtype=np.float64)

    for start in range(0, n_cells, chunk_size):
        end = min(start + chunk_size, n_cells)
        chunk = matrix.iloc[start:end]
        df_rnk = create_rankings(chunk, seed)
        rnk = df_rnk.iloc[:, df_rnk.columns.isin(geneset)]

        if rnk.empty or (float(len(rnk.columns)) / float(len(geneset))) < 0.80:
            print(
                f"Less than 80% of the genes in {geneset_name} are present in the "
                "expression matrix."
            )
        else:
            weights = np.array([1 for _ in geneset])
            auc_results[start:end] = aucs(rnk, len(df_rnk.columns), weights, auc_threshold)

    adata.obs[f'{geneset_name}_aucell'] = auc_results

def global_imports(modulename,shortname = None, asfunction = False):
    if shortname is None: 
        shortname = modulename
    if asfunction is False:
        globals()[shortname] = __import__(modulename)
    else:        
        globals()[shortname] = __import__(modulename)

def geneset_aucell(adata,geneset_name,geneset,AUC_threshold=0.01,seed=42):
        
    """
    Calculate the AUC-ell score for a given gene set.

    Arguments:
        adata: AnnData object
            Annotated data matrix containing gene expression data.
        geneset_name: str
            Name of the gene set.
        geneset: list of str
            List of gene symbols for the gene set.
        AUC_threshold: float, optional
            AUC threshold used to determine significant interactions (default is 0.01).
        seed: int, optional
            Seed used to initialize the random number generator (default is 42).

    Returns:
        None
           Adds a column to the 'obs' attribute of the adata object containing the AUC-ell score for the gene set.
    """

    check_ctxcore()
    global ctxcore_install
    if ctxcore_install==True:
        global aucs
        global GeneSignature
        from ctxcore.recovery import aucs
        from ctxcore.genesig import GeneSignature


    matrix = adata.X.copy()
    percentiles = derive_auc_threshold(matrix)
    auc_threshold = percentiles[AUC_threshold]

    np_rnk_sparse=fast_rank(matrix, seed= seed)

    rnk  = pd.DataFrame(np_rnk_sparse[:, np.where(adata.var_names.isin(geneset))[0]])

    if rnk.empty or (float(np_rnk_sparse.shape[1]) / float(len(geneset))) < 0.80:
        print(
            f"Less than 80% of the genes in {geneset_name} are present in the "
            "expression matrix."
        )
        adata.obs['{}_aucell'.format(geneset_name)]=np.zeros(shape=(rnk.shape[0]), dtype=np.float64)
    else:
        weights = np.array([1 for i in geneset])
        adata.obs['{}_aucell'.format(geneset_name)]=aucs(rnk,
        np_rnk_sparse.shape[1],weights,auc_threshold
        )
    
def pathway_aucell(adata,pathway_names,pathways_dict,AUC_threshold=0.01,seed=42):
    """
    Calculates the area under the curve (AUC) for a set of pathways in an AnnData object.

    Arguments:
        adata: AnnData object
            AnnData object containing the data.
        pathway_names: list of str
            Names of the pathways to analyze.
        pathways_dict: dict
            Dictionary containing the gene sets for each pathway.
        AUC_threshold: float, optional (default: 0.01)
            AUC threshold to use for determining significant gene-pathway associations.
        seed: int, optional (default: 42)
            Random seed for reproducibility.

    Returns:
        None
            The function modifies the `adata.obs` attribute of the input AnnData object.
    """


    check_ctxcore()
    global ctxcore_install
    if ctxcore_install==True:
        global aucs
        global GeneSignature
        from ctxcore.recovery import aucs
        from ctxcore.genesig import GeneSignature

    matrix = adata.X.copy()
    percentiles = derive_auc_threshold(matrix)
    auc_threshold = percentiles[AUC_threshold]

    np_rnk_sparse=fast_rank(matrix, seed= seed)

    for pathway_name in pathway_names:
        pathway_genes=pathways_dict[pathway_name]

        rnk  = pd.DataFrame(np_rnk_sparse[:, np.where(adata.var_names.isin(pathway_genes))[0]]).copy()
        if rnk.empty or (float(np_rnk_sparse.shape[1]) / float(len(pathway_genes))) < 0.80:
            print(
                f"Less than 80% of the genes in {pathway_name} are present in the "
                "expression matrix."
            )
            adata.obs['{}_aucell'.format(pathway_name)]=np.zeros(shape=(rnk.shape[0]), dtype=np.float64)
        else:
            weights = np.array([1 for i in pathway_genes])
            adata.obs['{}_aucell'.format(pathway_name)]=aucs(rnk,
            np_rnk_sparse.shape[1],weights,auc_threshold
            )
    add_reference(adata,'AUCell','pathway activity score with AUCell')
            
def pathway_aucell_tmp(adata, pathway_names, pathways_dict, AUC_threshold=0.01, seed=42, chunk_size=10000):
    """
    Calculates the area under the curve (AUC) for a set of pathways in an AnnData object.

    Arguments:
        adata: AnnData object
            AnnData object containing the data.
        pathway_names: list of str
            Names of the pathways to analyze.
        pathways_dict: dict
            Dictionary containing the gene sets for each pathway.
        AUC_threshold: float, optional (default: 0.01)
            AUC threshold to use for determining significant gene-pathway associations.
        seed: int, optional (default: 42)
            Random seed for reproducibility.
        chunk_size: int, optional (default: 10000)
            The number of cells to process in each chunk.

    Returns:
        None
            The function modifies the `adata.obs` attribute of the input AnnData object.
    """

    check_ctxcore()
    global ctxcore_install
    if ctxcore_install == True:
        global aucs
        global GeneSignature
        from ctxcore.recovery import aucs
        from ctxcore.genesig import GeneSignature

    matrix = adata.to_df()
    percentiles = derive_auc_threshold(matrix)
    auc_threshold = percentiles[AUC_threshold]
    
    n_cells = matrix.shape[0]
    
    for pathway_name in pathway_names:
        pathway_genes = pathways_dict[pathway_name]
        auc_results = np.zeros(n_cells, dtype=np.float64)
        
        for start in range(0, n_cells, chunk_size):
            end = min(start + chunk_size, n_cells)
            chunk = matrix.iloc[start:end]
            df_rnk = create_rankings(chunk, seed)
            
            rnk = df_rnk.iloc[:, df_rnk.columns.isin(pathway_genes)]
            if rnk.empty or (float(len(rnk.columns)) / float(len(pathway_genes))) < 0.80:
                print(
                    f"Less than 80% of the genes in {pathway_name} are present in the "
                    "expression matrix."
                )
            else:
                weights = np.array([1 for _ in pathway_genes])
                auc_results[start:end] = aucs(rnk, len(df_rnk.columns), weights, auc_threshold)

        adata.obs[f'{pathway_name}_aucell'] = auc_results
            
def pathway_aucell_enrichment(adata,pathways_dict,AUC_threshold=0.01,seed=42,num_workers=1):
    """
    Enriches cell annotations with pathway activity scores using the AUC-ell method.

    Arguments:
        adata: AnnData object
            AnnData object containing the expression matrix.
        pathways_dict: dict
            A dictionary where keys are pathway names and values are lists of genes associated with each pathway.
        AUC_threshold: float, optional
            The threshold for calculating the area under the curve (AUC) values using the AUC-ell method. The default is 0.01.
        seed: int, optional
            The seed to use for the random number generator. The default is 42.
        num_workers: int, optional
            The number of workers to use for parallel processing. The default is 1.

    Returns:
        adata_aucs: AnnData object
            AnnData object containing the pathway activity scores for each cell in the input AnnData object.
    """

    check_ctxcore()
    global ctxcore_install
    if ctxcore_install==True:
        global aucs
        global GeneSignature
        from ctxcore.recovery import aucs
        from ctxcore.genesig import GeneSignature
        
    test_gmt=[]
    for i in pathways_dict.keys():
        test_gmt.append(GeneSignature(name=i,gene2weight=dict(zip(pathways_dict[i],[1 for i in pathways_dict[i]]))))

    matrix = adata.to_df()
    percentiles = derive_auc_threshold(matrix)
    auc_threshold = percentiles[AUC_threshold]

    aucs_mtx = aucell(matrix, signatures=test_gmt, auc_threshold=auc_threshold, num_workers=num_workers)
    adata_aucs=anndata.AnnData(aucs_mtx)
    add_reference(adata,'AUCell','pathway activity score with AUCell')
    return adata_aucs

def pathway_aucell_enrichment_tmp(adata, pathways_dict, AUC_threshold=0.01, seed=42, 
                              num_workers=1, chunk_size=10000):
    """
    Enriches cell annotations with pathway activity scores using the AUC-ell method.

    Arguments:
        adata: AnnData object
            AnnData object containing the expression matrix.
        pathways_dict: dict
            A dictionary where keys are pathway names and values are lists of genes associated with each pathway.
        AUC_threshold: float, optional
            The threshold for calculating the area under the curve (AUC) values using the AUC-ell method. The default is 0.01.
        seed: int, optional
            The seed to use for the random number generator. The default is 42.
        num_workers: int, optional
            The number of workers to use for parallel processing. The default is 1.
        chunk_size: int, optional
            The number of cells to process in each chunk. The default is 10000.

    Returns:
        adata_aucs: AnnData object
            AnnData object containing the pathway activity scores for each cell in the input AnnData object.
    """
    from tqdm import tqdm
    check_ctxcore()
    global ctxcore_install
    if ctxcore_install == True:
        global aucs
        global GeneSignature
        from ctxcore.recovery import aucs
        from ctxcore.genesig import GeneSignature
        
    test_gmt = []
    for i in pathways_dict.keys():
        test_gmt.append(GeneSignature(name=i, gene2weight=dict(zip(pathways_dict[i], [1 for _ in pathways_dict[i]]))))

    matrix = adata.to_df()
    percentiles = derive_auc_threshold(matrix)
    auc_threshold = percentiles[AUC_threshold]

    # Process in chunks
    aucs_mtx_list = []
    n_cells = matrix.shape[0]
    for start in tqdm(range(0, n_cells, chunk_size)):
        end = min(start + chunk_size, n_cells)
        chunk = matrix.iloc[start:end]
        aucs_chunk = aucell(chunk, signatures=test_gmt, auc_threshold=auc_threshold, num_workers=num_workers)
        aucs_mtx_list.append(aucs_chunk)

    # Concatenate the results
    aucs_mtx = pd.concat(aucs_mtx_list)

    adata_aucs = anndata.AnnData(aucs_mtx)
    return adata_aucs


def pathway_enrichment(adata, pathways_dict,organism='Human',group_by='louvain', 
                       cutoff=0.05, logfc_threshold=2,pvalue_type='adjust',plot=True):
    
    """
    Perform pathway enrichment analysis on gene expression data.
    
    Arguments:
        adata: anndata.AnnData
            Annotated data matrix containing gene expression data.
        pathways_dict: dict
            A dictionary of gene sets with their gene members.
        organism: str, optional (default: 'Human')
            The organism to be used for the enrichment analysis. Can be either 'Human' or 'Mouse'.
        group_by: str, optional (default: 'louvain')
            The group label of the cells in adata.obs to perform the enrichment analysis on.
        cutoff: float, optional (default: 0.05)
            The adjusted p-value cutoff used to filter enriched pathways.
        logfc_threshold: float, optional (default: 2)
            The log2 fold change cutoff used to define differentially expressed genes.
        pvalue_type: str, optional (default: 'adjust')
            The type of p-value used for filtering enriched pathways. Can be either 'adjust' or 'raw'.
        plot: bool, optional (default: True)
            If True, generate a bar plot for each cluster of enriched pathways.
    
    Returns:
        enrich_res: pandas.DataFrame
            A pandas dataframe containing the enriched pathways information including term name, 
            p-value, adjusted p-value, overlap genes, odds ratio, log2 fold change, and cluster name.
    
    Examples:
        >>> # Perform pathway enrichment analysis on adata using pathways_dict
        >>> res = pathway_enrichment(adata, pathways_dict)
    
    Reference:
        The code for pathway_enrichment() function was adapted from scanpy workflows: 
        https://scanpy-tutorials.readthedocs.io/en/latest/pbmc3k.html#Gene-set-enrichment-analysis.
    """

    #import gseapy as gp
    from ..externel.gseapy import enrichr
    df_list = []
    cluster_list = []
    celltypes = sorted(adata.obs[group_by].unique())

    for celltype in celltypes:
        degs = sc.get.rank_genes_groups_df(adata, group=celltype, key='rank_genes_groups', log2fc_min=logfc_threshold, 
                                    pval_cutoff=cutoff)['names'].squeeze()
        if isinstance(degs, str):
            degs = [degs.strip()]
        else:
            degs = degs.str.strip().tolist()
        
        if not degs:
            continue
        if (organism == 'Mouse') or (organism == 'mouse') or (organism == 'mm'):
            background='mmusculus_gene_ensembl'
        elif (organism == 'Human') or (organism == 'human') or (organism == 'hs'):
            background='hsapiens_gene_ensembl'
        else:
            background=adata.var.index.tolist()
        enr = enrichr(gene_list=degs,
                description='',
                gene_sets=pathways_dict,
                organism=organism,
                background=background,
                cutoff=0.5,
                )
        if (enr is not None) and hasattr(enr, 'res2d') and (enr.res2d.shape[0] > 0):
            df_list.append(enr.res2d)
            cluster_list.append(celltype)

    columns = ['Cluster', 'Gene_set', 'Term', 'Overlap', 'Odds Ratio','P-value', 'Adjusted P-value', 'Genes']

    df = pd.DataFrame(columns = columns)
    for cluster_ind, df_ in zip(cluster_list, df_list):
        df_ = df_[df_['Adjusted P-value'] <= cutoff]
        df_ = df_.assign(Cluster = cluster_ind)
        if (df_.shape[0] > 0):
            df = pd.concat([df, df_[columns]], sort=False)
            df_tmp = df_.loc[:, ['Term', 'Adjusted P-value']][:min(10, df_.shape[0])]
            df_tmp['Term'] = [x.split('(',1)[0] for x in df_tmp['Term']]
            df_tmp['-log_adj_p'] = - np.log10(df_tmp['Adjusted P-value'])
            df_tmp = df_tmp.sort_values(by='-log_adj_p', ascending=True)
            if plot==True:
                ax = df_tmp.plot.barh(y='-log_adj_p', x='Term', legend=False, grid=False, figsize=(12,4))
                ax.set_title('Cluster {}'.format(cluster_ind))
                ax.set_ylabel('')
                ax.set_xlabel('-log(Adjusted P-value)')
            #pp.savefig(ax.figure, bbox_inches='tight')
            #plt.close()
        else:
            print('No pathway with an adjusted P-value less than the cutoff (={}) for cluster {}'.format(cutoff, cluster_ind))
    
    if pvalue_type=='adjust':
        enrich_res=df[df['Adjusted P-value']<cutoff]
        enrich_res['logp']=-np.log(enrich_res['Adjusted P-value'])
    else:
        enrich_res=df[df['P-value']<cutoff]
        enrich_res['logp']=-np.log(enrich_res['P-value'])
    enrich_res['logc']=np.log(enrich_res['Odds Ratio'])
    enrich_res['num']=[int(i.split('/')[0]) for i in enrich_res['Overlap']]
    enrich_res['fraction']=[int(i.split('/')[0])/int(i.split('/')[1]) for i in enrich_res['Overlap']]
    add_reference(adata,'GSEApy','pathway enrichment analysis with gseapy')
    
    return enrich_res

def pathway_enrichment_plot(enrich_res,term_num=5,return_table=False,figsize=(3,10),plot_title='',**kwds):
    
    """
    Visualize the pathway enrichment analysis results as a heatmap.

    Parameters:
        enrich_res: pandas DataFrame
            The output from the pathway_enrichment() function.
        term_num: int, optional
            The number of enriched terms to display for each cluster. Default is 5.
        return_table: bool, optional
            Whether to return the heatmap table as a DataFrame. Default is False.
        figsize: tuple, optional
            The size of the plot. Default is (3,10).
        plot_title: str, optional
            The title of the plot. Default is an empty string.
        **kwds: optional
            Other keyword arguments to pass to the seaborn heatmap function.

    Returns:
        ax: matplotlib Axes object
            The heatmap plot.

    Examples:
        >>> res = pathway_enrichment(adata, pathways_dict)
        >>> pathway_enrichment_plot(res, term_num=10, return_table=True, figsize=(6,12), cmap='Blues')
    """

    celltypes=enrich_res['Cluster'].unique()
    plot_heatmap=pd.DataFrame(index=celltypes)
    for celltype in celltypes:
        res_test=enrich_res.loc[enrich_res['Cluster']==celltype].iloc[:term_num]
        plot_heatmap_test=pd.DataFrame(res_test[['Term','logp']])
        plot_heatmap_test=plot_heatmap_test.set_index(plot_heatmap_test['Term'])
        del plot_heatmap_test['Term']
        #plot_heatmap[plot_heatmap_test.index]=0
        for i in plot_heatmap_test.index:
            if len(enrich_res.loc[(enrich_res['Cluster']==celltype)&(enrich_res['Term']==i),'logp'].values)!=0:
                plot_heatmap.loc[celltype,i]=enrich_res.loc[(enrich_res['Cluster']==celltype)&(enrich_res['Term']==i),'logp'].values[0]
            else:
                plot_heatmap.loc[celltype,i]=0
    plot_heatmap.fillna(0,inplace=True)
    if return_table==True:
        return plot_heatmap
    
    fig,ax=plt.subplots(figsize=figsize)
    axr=sns.heatmap(plot_heatmap.T,ax=ax,**kwds)
    axr.tick_params(right=True, top=False,left=False, labelright=True, labelleft=False,labeltop=False,rotation=0)
    ax.set_title(plot_title,fontsize=12)
    ax.set_yticklabels(ax.get_yticklabels(), rotation = 0, fontsize = 10)
    ax.set_xticklabels(ax.get_xticklabels(), rotation = 90, fontsize = 10)

    
    return ax






import numpy as np
import scanpy as sc
import squidpy as sq
import pandas as pd    


def get_gaston_input_adata(data_folder, get_rgb=False, spot_umi_threshold=50):
    adata=sq.read.visium(data_folder)
    adata.obsm['spatial']=adata.obsm['spatial'].astype(float)
    sc.pp.filter_cells(adata, min_counts=spot_umi_threshold)

    gene_labels=adata.var.index.to_numpy()
    counts_mat=adata.X
    coords_mat=np.array(adata.obsm['spatial'])

    if not get_rgb:
        return counts_mat, coords_mat, gene_labels

    library_id = list(adata.uns['spatial'].keys())[0] # adata2.uns['spatial'] should have only one key
    scale=adata.uns['spatial'][library_id]['scalefactors']['tissue_hires_scalef']
    img = sq.im.ImageContainer(adata.uns['spatial'][library_id]['images']['hires'],
                               scale=scale,
                               layer="img1")
    print('calculating RGB')
    sq.im.calculate_image_features(adata, img, features="summary", key_added="features")
    columns = ['summary_ch-0_mean', 'summary_ch-1_mean', 'summary_ch-2_mean']
    RGB_mean = adata.obsm["features"][columns]
    RGB_mean = RGB_mean / 255
    
    # RGB_mean = RGB_mean[RGB_mean.index.isin(df_pos['barcode'])]

    return counts_mat, coords_mat, gene_labels, RGB_mean.to_numpy()

def get_top_pearson_residuals(num_pcs, counts_mat, coords_mat, gene_labels=None, n_top_genes=5000, clip=0.01):
    df=pd.DataFrame(counts_mat, columns=gene_labels)
    adata=sc.AnnData(df)
    adata.obsm["coords"] = coords_mat
    
    sc.experimental.pp.highly_variable_genes(
            adata, flavor="pearson_residuals", n_top_genes=n_top_genes
    )
    
    adata = adata[:, adata.var["highly_variable"]]
    adata.layers["raw"] = adata.X.copy()
    adata.layers["sqrt_norm"] = np.sqrt(
        sc.pp.normalize_total(adata, inplace=False)["X"]
    )
    
    theta=np.Inf
    sc.experimental.pp.normalize_pearson_residuals(adata, clip=clip, theta=theta)
    sc.pp.pca(adata, n_comps=num_pcs)
    return adata.obsm['X_pca']

def get_gaston_input_xenium(folder,filter_zero_cells=False):
    adata = sc.read_10x_h5(filename=folder+'cell_feature_matrix.h5')

    df = pd.read_csv(folder+"cells.csv")
    df.set_index(adata.obs_names, inplace=True)
    adata.obs = df.copy()

    adata.obsm["spatial"] = adata.obs[["x_centroid", "y_centroid"]].copy().to_numpy()

    counts_mat=np.array(adata.X.todense())
    coords_mat=adata.obs[["x_centroid", "y_centroid"]].to_numpy()

    cts=pd.read_csv(folder+'Cell_Barcode_Type_Matrices.csv')
    cts=np.array(cts['Cluster'])

    gene_labels=np.array(adata.var.index)

    if filter_zero_cells:
        inds_to_keep=np.where(np.sum(counts_mat,1)>0)[0]
        counts_mat=counts_mat[inds_to_keep,:]
        coords_mat=coords_mat[inds_to_keep,:]
        cts=cts[inds_to_keep]

    return counts_mat, coords_mat, gene_labels, cts


"""Module providing a encapsulation of pySTAGATE."""
import torch
from torch_geometric.loader import DataLoader
from tqdm import tqdm
import torch.nn.functional as F
from ..externel.STAGATE_pyG import Batch_Data,Cal_Spatial_Net,Transfer_pytorch_Data,Stats_Spatial_Net,STAGATE
import scanpy as sc
from anndata import AnnData
import numpy as np
import os
from scipy.sparse import csr_matrix
from .._settings import add_reference

class pySTAGATE:
    """Class representing the object of pySTAGATE."""

    def __init__(self,
                 adata: AnnData,
                 num_batch_x,
                 num_batch_y,
                 spatial_key: list = ['X','Y'],
                 batch_size: int = 1,
                rad_cutoff: int = 200,
                num_epoch: int = 1000,
                lr: float = 0.001,
                weight_decay: float = 1e-4,
                hidden_dims: list = [512, 30],
                device: str = 'cuda:0')-> None:
        # Initialize device
        device = torch.device(device if torch.cuda.is_available() else 'cpu')
        self.device=device

        # Create batches
        batch_list = Batch_Data(adata, num_batch_x=num_batch_x, num_batch_y=num_batch_y,
                                    spatial_key=spatial_key, plot_Stats=True)
        for temp_adata in batch_list:
            Cal_Spatial_Net(temp_adata, rad_cutoff=rad_cutoff)

        # Transfer to PyTorch data format
        #device = torch.device('cuda:0' if torch.cuda.is_available() else 'cpu')
        data_list = [Transfer_pytorch_Data(adata) for adata in batch_list]
        for temp in data_list:
            temp.to(device)

        Cal_Spatial_Net(adata, rad_cutoff=rad_cutoff)
        data = Transfer_pytorch_Data(adata)
        Stats_Spatial_Net(adata)

        # batch_size=1 or 2
        self.loader = DataLoader(data_list, batch_size=batch_size, shuffle=True)

        # hyper-parameters
        self.num_epoch = num_epoch
        self.lr=lr
        self.weight_decay=weight_decay
        self.hidden_dims = hidden_dims
        self.adata=adata
        self.data=data

        # Model and optimizer
        self.model = STAGATE(hidden_dims = [data_list[0].x.shape[1]]+self.hidden_dims).to(device)
        self.optimizer = torch.optim.Adam(self.model.parameters(),
                                            lr=self.lr,
                                            weight_decay=self.weight_decay)

    def train(self):
        """Train the STAGATE model."""       
        for epoch in tqdm(range(1, self.num_epoch+1)):
            for batch in self.loader:
                self.model.train()
                self.optimizer.zero_grad()
                z, out = self.model(batch.x, batch.edge_index)
                loss = F.mse_loss(batch.x, out) #F.nll_loss(out[data.train_mask], data.y[data.train_mask])
                loss.backward()
                torch.nn.utils.clip_grad_norm_(self.model.parameters(), 5.)
                self.optimizer.step()
        # The total network
        self.data.to(self.device)

    def predicted(self):
        """
        Predict the STAGATE representation and ReX values for all cells.
        """
        self.model.eval()
        z, out = self.model(self.data.x, self.data.edge_index)

        stagate_rep = z.to('cpu').detach().numpy()
        self.adata.obsm['STAGATE'] = stagate_rep
        rex = out.to('cpu').detach().numpy()
        rex[rex<0] = 0
        self.adata.layers['STAGATE_ReX'] = rex

        print('The STAGATE representation values are stored in adata.obsm["STAGATE"].')
        print('The rex values are stored in adata.layers["STAGATE_ReX"].')

    def cal_pSM(self,n_neighbors:int=20,resolution:int=1,
                       max_cell_for_subsampling:int=5000,
                       psm_key='pSM_STAGATE'):
        """
        Calculate the pseudo-spatial map using diffusion pseudotime (DPT) algorithm.

        Parameters
        ----------
        n_neighbors: int
            Number of neighbors for constructing the kNN graph.
        resolution: float
            Resolution for clustering.
        max_cell_for_subsampling: int
            Maximum number of cells for subsampling. 
            If the number of cells is larger than this value, the subsampling will be performed.

        Returns
        -------
        pSM_values: numpy.ndarray
            The pseudo-spatial map values.
        
        """

        from scipy.spatial import distance_matrix
        import numpy as np

        sc.pp.neighbors(self.adata, n_neighbors=n_neighbors,
               use_rep='STAGATE')
        sc.tl.umap(self.adata)
        sc.tl.leiden(self.adata, resolution=resolution)
        sc.tl.paga(self.adata)
       # max_cell_for_subsampling = max_cell_for_subsampling
        if self.adata.shape[0] < max_cell_for_subsampling:
            sub_adata_x = self.adata.obsm['STAGATE']
        else:
            indices = np.arange(self.adata.shape[0])
            selected_ind = np.random.choice(indices, max_cell_for_subsampling, False)
            sub_adata_x = self.adata[selected_ind, :].obsm['STAGATE']

        sum_dists = distance_matrix(sub_adata_x, sub_adata_x).sum(axis=1)
        self.adata.uns['iroot'] = np.argmax(sum_dists)
        sc.tl.diffmap(self.adata)
        sc.tl.dpt(self.adata)
        self.adata.obs.rename({"dpt_pseudotime": psm_key}, axis=1, inplace=True)
        print(f'The pseudo-spatial map values are stored in adata.obs["{psm_key}"].')

        psm_values = self.adata.obs[psm_key].to_numpy()
        return psm_values
        # End-of-file (EOF)


def clusters(adata,
             methods,
             methods_kwargs,
             batch_key=None,
             lognorm=50*1e4,
             ):
    """
    This function is used to cluster the spot in spatial RNA-seq data.
    
    """
    from scipy.sparse import issparse
    from ..externel.GraphST import GraphST

    for method in methods:
        if method=='STAGATE':
            print('The STAGATE method is used to cluster the spatial data.')
            adata_copy=adata.copy()
            if issparse(adata_copy.obsm['spatial']):
                adata_copy.obsm['spatial']=adata_copy.obsm['spatial'].toarray()
            adata_copy.obs['X']=adata_copy.obsm['spatial'][:,0]
            adata_copy.obs['Y']=adata_copy.obsm['spatial'][:,1]
            if 'STAGATE' not in methods_kwargs:
                methods_kwargs['STAGATE']={'num_batch_x':3,'num_batch_y':2,
                                           'spatial_key':['X','Y'],'rad_cutoff':200,'num_epoch':1000,'lr':0.001,
                                           'weight_decay':1e-4,'hidden_dims':[512, 30],'device':'cuda:0',
                                           'n_top_genes':2000}
            if 'n_top_genes' in methods_kwargs['STAGATE']:
                sc.pp.highly_variable_genes(adata_copy, flavor="seurat_v3", 
                                            n_top_genes=methods_kwargs['STAGATE']['n_top_genes'])
                del methods_kwargs['STAGATE']['n_top_genes']

            STA_obj=pySTAGATE(adata_copy,**methods_kwargs['STAGATE'])
            STA_obj.train()
            STA_obj.predicted()
            adata.obsm['STAGATE']=adata_copy.obsm['STAGATE']
            adata.layers['STAGATE_ReX']=adata_copy.layers['STAGATE_ReX']
            print(f'The STAGATE embedding are stored in adata.obsm["STAGATE"].\nShape: {adata.obsm["STAGATE"].shape}')
            add_reference(adata,'STAGATE','clustering with STAGATE')

        elif method=='GraphST':
            print('The GraphST method is used to embed the spatial data.')
            adata_copy=adata.copy()
            if 'GraphST' not in methods_kwargs:
                methods_kwargs['GraphST']={'device':'cuda:0','n_pcs':30,
                                           }
            if 'counts' not in adata_copy.layers:
                from ..pp import recover_counts
                if issparse(adata_copy.X):
                    pass 
                else:
                    adata_copy.X=csr_matrix(adata_copy.X)
                print('Recover the counts matrix from log-normalized data.')
                X_counts_recovered, size_factors_sub=recover_counts(adata_copy.X, lognorm, 
                                                                    lognorm*10,  log_base=None, 
                                                          chunk_size=10000)
                adata_copy.X=X_counts_recovered
            else:
                adata_copy.X=adata_copy.layers['counts']
            
            GRT_obj=GraphST(adata_copy, **methods_kwargs['GraphST'])
            adata_copy=GRT_obj.train()
            adata.obsm['GraphST_embedding']=adata_copy.obsm['GraphST_embedding']
            adata.obsm['graphst|original|X_pca']=adata_copy.obsm['graphst|original|X_pca']

            print(f'The GraphST embedding are stored in adata.obsm["GraphST_embedding"]. \nShape: {adata.obsm["GraphST_embedding"].shape}')
            add_reference(adata,'GraphST','clustering with GraphST')

        elif method=='CAST':
            print('The CAST method is used to embed the spatial data.')
            if issparse(adata.obsm['spatial']):
                adata.obsm['spatial']=adata.obsm['spatial'].toarray()
            adata.obs['X']=adata.obsm['spatial'][:,0]
            adata.obs['Y']=adata.obsm['spatial'][:,1]

            if batch_key is None:
                adata.obs['CAST_sample']='sample1'
            else:
                adata.obs['CAST_sample']=adata.obs[batch_key]
            
            if 'CAST' not in methods_kwargs:
                methods_kwargs['CAST']={'output_path_t':'result/CAST_gas/output',
                                        'device':'cuda:0',
                                        'gpu_t':0}
            # Get the coordinates and expression data for each sample
            samples = np.unique(adata.obs['CAST_sample']) # used samples in adata
            coords_raw = {sample_t: np.array(adata.obs[['X','Y']])[adata.obs['CAST_sample'] == sample_t] for sample_t in samples}
            
            if ('norm_1e4' not in adata.layers) and ('counts' in adata.layers):
                adata.layers['norm_1e4'] = sc.pp.normalize_total(adata, target_sum=1e4, layer='counts',
                                                 inplace=False)['X'].toarray() # we use normalized counts for each cell as input gene expression
                exp_dict = {sample_t: adata[adata.obs['CAST_sample'] == sample_t].layers['norm_1e4'] for sample_t in samples}
            else:
                exp_dict = {sample_t: adata[adata.obs['CAST_sample'] == sample_t].X for sample_t in samples}
            
            output_path = methods_kwargs['CAST']['output_path_t']
            os.makedirs(output_path, exist_ok=True)

            from ..externel.CAST import CAST_MARK
            embed_dict = CAST_MARK(coords_raw,exp_dict,**methods_kwargs['CAST'])

            from tqdm import tqdm
            adata.obsm['X_cast']=np.zeros((adata.shape[0],512))
            import pandas as pd
            adata.obsm['X_cast']=pd.DataFrame(adata.obsm['X_cast'],index=adata.obs.index)
            for key in tqdm(embed_dict.keys()):
                adata.obsm['X_cast'].loc[adata.obs['CAST_sample']==key]+=embed_dict[key].cpu().numpy()
            adata.obsm['X_cast']=adata.obsm['X_cast'].values
            print(f'The CAST embedding are stored in adata.obsm["X_cast"]. \nShape: {adata.obsm["X_cast"].shape}')
            add_reference(adata,'CAST','embedding with CAST')
        elif method=='BINARY':
            print('The BINARY method is used to embed the spatial data.')
            from ..externel import BINARY
            if batch_key is None:
                adata.obs['BINARY_sample']='sample1'
            else:
                adata.obs['BINARY_sample']=adata.obs[batch_key]
            if 'BINARY' not in methods_kwargs:
                methods_kwargs['BINARY']={'use_method':'KNN',
                                          'cutoff':6,
                                          'obs_key':'BINARY_sample',
                                          'use_list':None,
                                          'pos_weight':10,
                                          'device':'cuda:0',
                                          'hidden_dims':[512, 30],
                                            'n_epochs': 1000,
                                            'lr':  0.001,
                                            'key_added': 'BINARY',
                                            'gradient_clipping': 5,
                                            'weight_decay': 0.0001,
                                            'verbose': True,
                                            'random_seed':0,
                                            #'lognorm':50*1e4,
                                            'n_top_genes':2000}
            adata_copy = BINARY.clean_adata(adata, save_obs=[methods_kwargs['BINARY']['obs_key']])
            if 'counts' in adata.layers:
                adata_copy.X=adata[adata_copy.obs.index].layers['counts']
            if 'counts' not in adata_copy.layers:
                from ..pp import recover_counts
                if issparse(adata_copy.X):
                    pass 
                else:
                    adata_copy.X=csr_matrix(adata_copy.X)
                print('Recover the counts matrix from log-normalized data.')
                X_counts_recovered, size_factors_sub=recover_counts(adata_copy.X, lognorm, 
                                                                    lognorm*10,  log_base=None, 
                                                          chunk_size=10000)
                adata_copy.X=X_counts_recovered
            else:
                adata_copy.X=adata_copy.layers['counts']
                #recover_counts(adata_copy.X, mult_value, max_range, log_base=None, chunk_size=1000)
            adata_copy = BINARY.Count2Binary(adata_copy)
            if 'n_top_genes' in methods_kwargs['BINARY']:
                sc.pp.highly_variable_genes(adata_copy, flavor="seurat_v3", n_top_genes=methods_kwargs['BINARY']['n_top_genes'])

            BINARY.Mutil_Construct_Spatial_Graph(adata_copy,
                                           use_method=methods_kwargs['BINARY']['use_method'],
                                           cutoff=methods_kwargs['BINARY']['cutoff'],
                                           obs_key=methods_kwargs['BINARY']['obs_key'],
                                           use_list=methods_kwargs['BINARY']['use_list'],
                                     )
            adata_copy = BINARY.train_BINARY(adata_copy,pos_weight=methods_kwargs['BINARY']['pos_weight'],
                                            device=methods_kwargs['BINARY']['device'],
                                            hidden_dims=methods_kwargs['BINARY']['hidden_dims'],
                                            n_epochs=methods_kwargs['BINARY']['n_epochs'],
                                            lr=methods_kwargs['BINARY']['lr'],
                                            key_added=methods_kwargs['BINARY']['key_added'],
                                            gradient_clipping=methods_kwargs['BINARY']['gradient_clipping'],
                                            weight_decay=methods_kwargs['BINARY']['weight_decay'],
                                            verbose=methods_kwargs['BINARY']['verbose'],
                                            random_seed=methods_kwargs['BINARY']['random_seed'])
            adata.obsm['BINARY']=adata_copy.obsm['BINARY']
            adata.uns['Spatial_Graph']=adata_copy.uns['Spatial_Graph']
            print(f'The binary embedding are stored in adata.obsm["BINARY"]. \nShape: {adata.obsm["BINARY"].shape}')
            add_reference(adata,'BINARY','clustering with BINARY')
        else:
            print(f'The method {method} is not supported.')
    return adata

def merge_cluster(adata,groupby='mclust',use_rep='STAGATE',
                  threshold=0.05,plot=True,start_idx=0,**kwargs):
    sc.tl.dendrogram(adata,groupby=groupby,use_rep=use_rep)
    import numpy as np
    from scipy.cluster.hierarchy import fcluster

    # 你的链接矩阵
    linkage_matrix = adata.uns[f'dendrogram_{groupby}']['linkage']

    # 选择一个阈值来确定簇
    #threshold = 0.05  # 这个值需要根据具体情况调整

    # 使用fcluster来合并类别
    clusters = fcluster(linkage_matrix, threshold, criterion='distance')
    
    # 创建字典
    cluster_dict = {}
    for idx, cluster_id in enumerate(clusters):
        key = f'c{cluster_id}'
        if key not in cluster_dict:
            cluster_dict[key] = []
        cluster_dict[key].append(idx+start_idx)
    
    reversed_dict = {}
    for key, values in cluster_dict.items():
        for value in values:
            reversed_dict[str(value)] = key
    
    #adata.obs['mclust_tree']=adata.obs['mclust'].map(reversed_dict)
    adata.obs[groupby]=adata.obs[groupby].astype(str)
    adata.obs[f'{groupby}_tree']=adata.obs[groupby].map(reversed_dict)
    print(f'The merged cluster information is stored in adata.obs["{groupby}_tree"].')
    if plot:
        ax=sc.pl.dendrogram(adata,groupby=groupby,show=False,**kwargs)
        ax.plot((ax.get_xticks().min(),ax.get_xticks().max()),(threshold,threshold))
    return reversed_dict


import matplotlib.pyplot as plt
import networkx as nx
import seaborn as sns
import scanpy as sc
import numpy as np 
import pandas as pd
from typing import Union, Sequence

palette_network = list(sns.color_palette("tab20") \
                       + sns.color_palette("tab20b") \
                       + sns.color_palette("tab20c")\
                       + sns.color_palette("Set1")\
                       + sns.color_palette("Set2"))

def subset_for_flow_type(adata: sc.AnnData,
                         var_type: str = 'all',
                         flowsig_expr_key: str = 'X_flow',
                         flowsig_network_key: str = 'flowsig_network'):
    
    var_types = ['all', 'inflow', 'module', 'outflow']

    if var_type not in var_types:
        ValueError("Need to specify var_type as one of the following: %s"  % var_types)

    X_flow = adata.obsm[flowsig_expr_key]
    adata_subset = sc.AnnData(X=X_flow)
    adata_subset.obs = adata.obs
    adata_subset.var = adata.uns[flowsig_network_key]['flow_var_info']

    if var_type != 'all':

        adata_subset = adata_subset[:, adata_subset.var['Type'].isin([var_type])]

    return adata_subset

def label_point(x, y, val, ax):

    a = pd.concat({'x': x, 'y': y, 'val': val}, axis=1)

    for i, point in a.iterrows():

        ax.text(point['x']+0.3, point['y'] + 0.1, str(point['val']), fontdict={'size':12.0})


def plot_differentially_flowing_signals(adata: sc.AnnData,
                                        condition_key: str,
                                        pert_key: str,
                                        var_type: str = 'all',
                                        flowsig_expr_key: str = 'X_flow',
                                        flowsig_network_key: str = 'flowsig_network',
                                        qval_threshold: float = 0.05,
                                        logfc_threshold: float = 0.5,
                                        label_lowqval: bool = False):

    var_types = ['all', 'inflow', 'module', 'outflow']

    adata_subset = subset_for_flow_type(adata = adata,
                                        var_type = var_type,
                                        flowsig_expr_key = flowsig_expr_key,
                                        flowsig_network_key = flowsig_network_key)
    
    if condition_key not in adata_subset.uns:
        adata_subset.uns['log1p'] = {'base': None}
        sc.tl.rank_genes_groups(adata_subset, key_added=condition_key, groupby=condition_key, method='wilcoxon')

    result = sc.get.rank_genes_groups_df(adata_subset, group=pert_key, key=condition_key).copy()
    result["-logQ"] = -np.log(result["pvals"].astype("float"))
    lowqval_de_received = result.loc[(abs(result["logfoldchanges"]) > logfc_threshold)&(abs(result["pvals_adj"]) < qval_threshold)]
    other_de_received = result.loc[(abs(result["logfoldchanges"]) <= logfc_threshold)|(abs(result["pvals_adj"]) >= qval_threshold)]

    fig, ax = plt.subplots()
    sns.regplot(
        x=other_de_received["logfoldchanges"],
        y=other_de_received["-logQ"],
        fit_reg=False,
        scatter_kws={"s": 50},
    )
    sns.regplot(
        x=lowqval_de_received["logfoldchanges"],
        y=lowqval_de_received["-logQ"],
        fit_reg=False,
        scatter_kws={"s": 50},
    )
    ax.set_xlabel("log2 FC")
    ax.set_ylabel("-log Q-value")

    if label_lowqval:
        label_point(lowqval_de_received['logfoldchanges'],
                    lowqval_de_received['-logQ'],
                    lowqval_de_received['names'],
                    plt.gca())  


def plot_intercellular_flows_from_inflows(adata: sc.AnnData,
                                          inflow_vars: Union[str, Sequence[str]],
                                          flow_network: nx.DiGraph,
                                          flowsig_network_key: str = 'flowsig_network',
                                          align_mode: str = 'horizontal',
                                          width_scale: float = 2.0,
                                          x_margin_offset: float = 0.3,
                                          y_margin_offset: float = 0.0):
    
    flow_vars = adata.uns[flowsig_network_key]['flow_var_info'].index.tolist()
    flow_var_info = adata.uns[flowsig_network_key]['flow_var_info']

    inflow_vars = []
    all_gems = [node for node in flow_vars\
                        if flow_var_info.loc[node]['Type'] == 'module']
    all_outflow_vars = [node for node in flow_vars\
                        if flow_var_info.loc[node]['Type'] == 'outflow']

    node_colours = {}
    count = 0

    for inflow_var in inflow_vars:
        node_colours[inflow_var] = palette_network[count]
        count += 1

    associated_gems = sorted([edge[1] for edge in flow_network.edges() if edge[0] in inflow_vars and edge[1] in all_gems])
    for gem in associated_gems:
        node_colours[gem] = palette_network[count]
        count += 1

    downstream_outflows = []
    for gem in associated_gems:

        outflow_successors = [node for node in flow_network.successors(gem) if node in all_outflow_vars]
        
        for outflow_var in outflow_successors:
            if outflow_var not in downstream_outflows:
                downstream_outflows.append(outflow_var)            

    downstream_outflows = sorted(downstream_outflows)

    for outflow_var in downstream_outflows:
        node_colours[outflow_var] = palette_network[count]
        count += 1

    resultant_pattern_graph = flow_network.subgraph(inflow_vars + associated_gems + downstream_outflows)

    for node in resultant_pattern_graph.nodes():

        if align_mode == 'horizontal':

            if node in inflow_vars:
                resultant_pattern_graph.nodes[node]['level'] = 3

            elif node in associated_gems:
                resultant_pattern_graph.nodes[node]['level'] = 2

            else: # Downstream outflow vars
                resultant_pattern_graph.nodes[node]['level'] = 1

        else: # Otherwise we're going veritcally

            if node in inflow_vars:
                resultant_pattern_graph.nodes[node]['level'] = 1

            elif node in associated_gems:
                resultant_pattern_graph.nodes[node]['level'] = 2

            else:
                resultant_pattern_graph.nodes[node]['level'] = 3

    resultant_pattern_graph_pos = nx.multipartite_layout(resultant_pattern_graph, subset_key='level', align=align_mode, scale=2.0)


    resultant_pattern_graph_edge_colours = [node_colours[edge[0]] for edge in resultant_pattern_graph.edges()]
    edge_widths = [width_scale*resultant_pattern_graph[u][v]['weight'] for u,v in resultant_pattern_graph.edges()]
    nx.draw_networkx_edges(resultant_pattern_graph, resultant_pattern_graph_pos, edge_color=resultant_pattern_graph_edge_colours, width=edge_widths, alpha=0.75, connectionstyle="arc3,rad=0.2")
    nx.draw_networkx_labels(resultant_pattern_graph, resultant_pattern_graph_pos, font_size=14, font_family='Arial');
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=inflow_vars, node_color=[node_colours[node] for node in inflow_vars], linewidths=1.0, edgecolors='black', node_size=500)
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=associated_gems, node_color=[node_colours[node] for node in associated_gems], node_shape='v', node_size=100)
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=downstream_outflows, node_color=[node_colours[node] for node in downstream_outflows], node_shape='s', node_size=100)

    plt.tight_layout()
    plt.margins(x=x_margin_offset)
    plt.margins(y=y_margin_offset)
    plt.box(False)
    plt.show()

def plot_intercellular_flows_from_gems(adata: sc.AnnData,
                                          gem_vars: Union[str, Sequence[str]],
                                          flow_network: nx.DiGraph,
                                          flowsig_network_key: str = 'flowsig_network',
                                          align_mode: str = 'horizontal',
                                          width_scale: float = 2.0,
                                          x_margin_offset: float = 0.3,
                                          y_margin_offset: float = 0.0):
    
    flow_vars = adata.uns[flowsig_network_key]['flow_var_info'].index.tolist()
    flow_var_info = adata.uns[flowsig_network_key]['flow_var_info']

    all_inflow_vars = [node for node in flow_vars\
                        if flow_var_info.loc[node]['Type'] == 'inflow']
    all_outflow_vars = [node for node in flow_vars\
                        if flow_var_info.loc[node]['Type'] == 'outflow']

    node_colours = {}
    count = 0

    upstream_inflows = []
    for gem in gem_vars:

        inflow_predecessors = [node for node in flow_network.predecessors(gem) if node in all_inflow_vars]
        
        for inflow_var in inflow_predecessors:
            if inflow_var not in upstream_inflows:
                upstream_inflows.append(inflow_var)            

    upstream_inflows = sorted(upstream_inflows)

    for inflow_var in upstream_inflows:
        node_colours[inflow_var] = palette_network[count]
        count += 1

    downstream_outflows = []
    for gem in gem_vars:

        outflow_successors = [node for node in flow_network.successors(gem) if node in all_outflow_vars]
        
        for outflow_var in outflow_successors:
            if outflow_var not in downstream_outflows:
                downstream_outflows.append(outflow_var)            

    downstream_outflows = sorted(downstream_outflows)

    for outflow_var in downstream_outflows:
        node_colours[outflow_var] = palette_network[count]
        count += 1

    resultant_pattern_graph = flow_network.subgraph(upstream_inflows + gem_vars + downstream_outflows)

    for node in resultant_pattern_graph.nodes():

        if align_mode == 'horizontal':

            if node in upstream_inflows:
                resultant_pattern_graph.nodes[node]['level'] = 3

            elif node in gem_vars:
                resultant_pattern_graph.nodes[node]['level'] = 2

            else: # Downstream outflow vars
                resultant_pattern_graph.nodes[node]['level'] = 1

        else: # Otherwise we're going veritcally

            if node in upstream_inflows:
                resultant_pattern_graph.nodes[node]['level'] = 1

            elif node in gem_vars:
                resultant_pattern_graph.nodes[node]['level'] = 2

            else:
                resultant_pattern_graph.nodes[node]['level'] = 3

    resultant_pattern_graph_pos = nx.multipartite_layout(resultant_pattern_graph, subset_key='level', align=align_mode, scale=2.0)

    resultant_pattern_graph_edge_colours = [node_colours[edge[0]] for edge in resultant_pattern_graph.edges()]
    edge_widths = [width_scale*resultant_pattern_graph[u][v]['weight'] for u,v in resultant_pattern_graph.edges()]
    nx.draw_networkx_edges(resultant_pattern_graph, resultant_pattern_graph_pos, edge_color=resultant_pattern_graph_edge_colours, width=edge_widths, alpha=0.75, connectionstyle="arc3,rad=0.2")
    nx.draw_networkx_labels(resultant_pattern_graph, resultant_pattern_graph_pos, font_size=14, font_family='Arial');
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=upstream_inflows, node_color=[node_colours[node] for node in upstream_inflows], linewidths=1.0, edgecolors='black', node_size=100)
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=gem_vars, node_color=[node_colours[node] for node in gem_vars], node_shape='v', node_size=500)
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=downstream_outflows, node_color=[node_colours[node] for node in downstream_outflows], node_shape='s', node_size=100)

    plt.tight_layout()
    plt.margins(x=x_margin_offset)
    plt.margins(y=y_margin_offset)
    plt.box(False)
    plt.show()

def plot_intercellular_flows_from_outflows(adata: sc.AnnData,
                                          outflow_vars: Union[str, Sequence[str]],
                                          flow_network: nx.DiGraph,
                                          flowsig_network_key: str = 'flowsig_network',
                                          align_mode: str = 'horizontal',
                                          width_scale: float = 2.0,
                                          x_margin_offset: float = 0.3,
                                          y_margin_offset: float = 0.0):
    
    flow_vars = adata.uns[flowsig_network_key]['flow_var_info'].index.tolist()
    flow_var_info = adata.uns[flowsig_network_key]['flow_var_info']

    all_inflow_vars = [node for node in flow_vars\
                        if flow_var_info.loc[node]['Type'] == 'inflow']
    all_gems = [node for node in flow_vars\
                        if flow_var_info.loc[node]['Type'] == 'module']

    node_colours = {}
    count = 0

    for outflow_var in outflow_vars:
        node_colours[outflow_var] = palette_network[count]
        count += 1

    associated_gems = sorted([edge[0] for edge in flow_network.edges() if edge[1] in outflow_vars and edge[0] in all_gems])
    for gem in associated_gems:
        node_colours[gem] = palette_network[count]
        count += 1

    upstream_inflows = []
    for gem in associated_gems:

        inflow_predecessors = [node for node in flow_network.predecessors(gem) if node in all_inflow_vars]
        
        for inflow_var in inflow_predecessors:
            if inflow_var not in upstream_inflows:
                upstream_inflows.append(inflow_var)            

    upstream_inflows = sorted(upstream_inflows)

    for inflow_var in upstream_inflows:
        node_colours[outflow_var] = palette_network[count]
        count += 1

    resultant_pattern_graph = flow_network.subgraph(upstream_inflows + associated_gems + outflow_vars)

    for node in resultant_pattern_graph.nodes():

        if align_mode == 'horizontal':

            if node in upstream_inflows:
                resultant_pattern_graph.nodes[node]['level'] = 3

            elif node in associated_gems:
                resultant_pattern_graph.nodes[node]['level'] = 2

            else: # Downstream outflow vars
                resultant_pattern_graph.nodes[node]['level'] = 1

        else: # Otherwise we're going veritcally

            if node in upstream_inflows:
                resultant_pattern_graph.nodes[node]['level'] = 1

            elif node in associated_gems:
                resultant_pattern_graph.nodes[node]['level'] = 2

            else:
                resultant_pattern_graph.nodes[node]['level'] = 3

    resultant_pattern_graph_pos = nx.multipartite_layout(resultant_pattern_graph, subset_key='level', align=align_mode, scale=2.0)

    resultant_pattern_graph_edge_colours = [node_colours[edge[0]] for edge in resultant_pattern_graph.edges()]
    edge_widths = [width_scale*resultant_pattern_graph[u][v]['weight'] for u,v in resultant_pattern_graph.edges()]
    nx.draw_networkx_edges(resultant_pattern_graph, resultant_pattern_graph_pos, edge_color=resultant_pattern_graph_edge_colours, width=edge_widths, alpha=0.75, connectionstyle="arc3,rad=0.2")
    nx.draw_networkx_labels(resultant_pattern_graph, resultant_pattern_graph_pos, font_size=14, font_family='Arial');
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=upstream_inflows, node_color=[node_colours[node] for node in upstream_inflows], linewidths=1.0, edgecolors='black', node_size=100)
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=associated_gems, node_color=[node_colours[node] for node in associated_gems], node_shape='v', node_size=100)
    nx.draw_networkx_nodes(resultant_pattern_graph, resultant_pattern_graph_pos, nodelist=outflow_vars, node_color=[node_colours[node] for node in outflow_vars], node_shape='s', node_size=500)

    plt.tight_layout()
    plt.margins(x=x_margin_offset)
    plt.margins(y=y_margin_offset)
    plt.box(False)
    plt.show()



# This file allows separating the most CPU intensive routines from the
# main code.  This allows them to be optimized with Cython.  If you
# don't have Cython, this will run normally.  However, if you use
# Cython, you'll get speed boosts from 10-100x automatically.
#
# THE ONLY CATCH IS THAT IF YOU MODIFY THIS FILE, YOU MUST ALSO MODIFY
# fa2util.pxd TO REFLECT ANY CHANGES IN FUNCTION DEFINITIONS!
#
# Copyright (C) 2017 Bhargav Chippada <bhargavchippada19@gmail.com>
#
# Available under the GPLv3

from math import sqrt
from libc.math cimport sqrt as c_sqrt, pow as c_pow

# Import Cython's features for faster execution
import cython
from cython.operator cimport dereference as deref, preincrement as inc

# This will substitute for the nLayout object
cdef class Node:
    def __init__(self):
        self.mass = 0.0
        self.old_dx = 0.0
        self.old_dy = 0.0
        self.dx = 0.0
        self.dy = 0.0
        self.x = 0.0
        self.y = 0.0


# This is not in the original java code, but it makes it easier to deal with edges
cdef class Edge:
    def __init__(self):
        self.node1 = -1
        self.node2 = -1
        self.weight = 0.0


# Here are some functions from ForceFactory.java
# =============================================

# Repulsion function.  `n1` and `n2` should be nodes.  This will
# adjust the dx and dy values of `n1`  `n2`
cdef void linRepulsion(Node n1, Node n2, double coefficient=0):
    cdef double xDist = n1.x - n2.x
    cdef double yDist = n1.y - n2.y
    cdef double distance2 = xDist * xDist + yDist * yDist  # Distance squared

    if distance2 > 0:
        cdef double factor = coefficient * n1.mass * n2.mass / distance2
        n1.dx += xDist * factor
        n1.dy += yDist * factor
        n2.dx -= xDist * factor
        n2.dy -= yDist * factor


# Repulsion function. 'n' is node and 'r' is region
cdef void linRepulsion_region(Node n, Region r, double coefficient=0):
    cdef double xDist = n.x - r.massCenterX
    cdef double yDist = n.y - r.massCenterY
    cdef double distance2 = xDist * xDist + yDist * yDist

    if distance2 > 0:
        cdef double factor = coefficient * n.mass * r.mass / distance2
        n.dx += xDist * factor
        n.dy += yDist * factor


# Gravity repulsion function.  For some reason, gravity was included
# within the linRepulsion function in the original gephi java code,
# which doesn't make any sense (considering a. gravity is unrelated to
# nodes repelling each other, and b. gravity is actually an
# attraction)
cdef void linGravity(Node n, double g):
    cdef double xDist = n.x
    cdef double yDist = n.y
    cdef double distance = c_sqrt(xDist * xDist + yDist * yDist)

    if distance > 0:
        cdef double factor = n.mass * g / distance
        n.dx -= xDist * factor
        n.dy -= yDist * factor


# Strong gravity force function. `n` should be a node, and `g`
# should be a constant by which to apply the force.
cdef void strongGravity(Node n, double g, double coefficient=0):
    cdef double xDist = n.x
    cdef double yDist = n.y

    if xDist != 0 and yDist != 0:
        cdef double factor = coefficient * n.mass * g
        n.dx -= xDist * factor
        n.dy -= yDist * factor


# Attraction function.  `n1` and `n2` should be nodes.  This will
# adjust the dx and dy values of `n1` and `n2`.  It does
# not return anything.
cpdef void linAttraction(Node n1, Node n2, double e, bint distributedAttraction, double coefficient=0):
    cdef double xDist = n1.x - n2.x
    cdef double yDist = n1.y - n2.y
    cdef double factor
    
    if not distributedAttraction:
        factor = -coefficient * e
    else:
        factor = -coefficient * e / n1.mass
        
    n1.dx += xDist * factor
    n1.dy += yDist * factor
    n2.dx -= xDist * factor
    n2.dy -= yDist * factor


# The following functions iterate through the nodes or edges and apply
# the forces directly to the node objects.  These iterations are here
# instead of the main file because Python is slow with loops.
cpdef void apply_repulsion(list nodes, double coefficient, list all_nodes=None):
    """Apply repulsion forces between nodes
    
    Args:
        nodes: List of nodes to calculate repulsion for
        coefficient: Repulsion coefficient
        all_nodes: If provided, calculate repulsion against these nodes, 
                  otherwise use nodes for both source and target
    """
    cdef int i, j
    cdef Node n1, n2
    
    if all_nodes is None:
        all_nodes = nodes
        
    for n1 in nodes:
        for n2 in all_nodes:
            if n1 is not n2:  # Skip self-repulsion
                linRepulsion(n1, n2, coefficient)


cpdef void apply_gravity(list nodes, double gravity, double scalingRatio, bint useStrongGravity=False):
    cdef Node n
    
    if not useStrongGravity:
        for n in nodes:
            linGravity(n, gravity)
    else:
        for n in nodes:
            strongGravity(n, gravity, scalingRatio)


cpdef void apply_attraction(list nodes, list edges, bint distributedAttraction, double coefficient, double edgeWeightInfluence):
    cdef Edge edge
    
    # Optimization, since usually edgeWeightInfluence is 0 or 1, and pow is slow
    if edgeWeightInfluence == 0:
        for edge in edges:
            linAttraction(nodes[edge.node1], nodes[edge.node2], 1, distributedAttraction, coefficient)
    elif edgeWeightInfluence == 1:
        for edge in edges:
            linAttraction(nodes[edge.node1], nodes[edge.node2], edge.weight, distributedAttraction, coefficient)
    else:
        for edge in edges:
            linAttraction(nodes[edge.node1], nodes[edge.node2], c_pow(edge.weight, edgeWeightInfluence),
                          distributedAttraction, coefficient)


# For Barnes Hut Optimization
cdef class Region:
    def __init__(self, nodes):
        self.mass = 0.0
        self.massCenterX = 0.0
        self.massCenterY = 0.0
        self.size = 0.0
        self.nodes = nodes
        self.subregions = []
        self.updateMassAndGeometry()

    cdef void updateMassAndGeometry(self):
        cdef double massSumX, massSumY, distance
        cdef Node n
        
        if len(self.nodes) > 1:
            self.mass = 0
            massSumX = 0
            massSumY = 0
            for n in self.nodes:
                self.mass += n.mass
                massSumX += n.x * n.mass
                massSumY += n.y * n.mass
            self.massCenterX = massSumX / self.mass
            self.massCenterY = massSumY / self.mass

            self.size = 0.0
            for n in self.nodes:
                distance = c_sqrt((n.x - self.massCenterX) ** 2 + (n.y - self.massCenterY) ** 2)
                self.size = max(self.size, 2 * distance)

    cpdef void buildSubRegions(self):
        cdef list topleftNodes, bottomleftNodes, toprightNodes, bottomrightNodes
        cdef Node n
        cdef Region subregion
        
        if len(self.nodes) > 1:
            topleftNodes = []
            bottomleftNodes = []
            toprightNodes = []
            bottomrightNodes = []
            # Optimization: The distribution of self.nodes into 
            # subregions now requires only one for loop. Removed 
            # topNodes and bottomNodes arrays: memory space saving.
            for n in self.nodes:
                if n.x < self.massCenterX:
                    if n.y < self.massCenterY:
                        bottomleftNodes.append(n)
                    else:
                        topleftNodes.append(n)
                else:
                    if n.y < self.massCenterY:
                        bottomrightNodes.append(n)
                    else:
                        toprightNodes.append(n)      

            if len(topleftNodes) > 0:
                if len(topleftNodes) < len(self.nodes):
                    subregion = Region(topleftNodes)
                    self.subregions.append(subregion)
                else:
                    for n in topleftNodes:
                        subregion = Region([n])
                        self.subregions.append(subregion)

            if len(bottomleftNodes) > 0:
                if len(bottomleftNodes) < len(self.nodes):
                    subregion = Region(bottomleftNodes)
                    self.subregions.append(subregion)
                else:
                    for n in bottomleftNodes:
                        subregion = Region([n])
                        self.subregions.append(subregion)

            if len(toprightNodes) > 0:
                if len(toprightNodes) < len(self.nodes):
                    subregion = Region(toprightNodes)
                    self.subregions.append(subregion)
                else:
                    for n in toprightNodes:
                        subregion = Region([n])
                        self.subregions.append(subregion)

            if len(bottomrightNodes) > 0:
                if len(bottomrightNodes) < len(self.nodes):
                    subregion = Region(bottomrightNodes)
                    self.subregions.append(subregion)
                else:
                    for n in bottomrightNodes:
                        subregion = Region([n])
                        self.subregions.append(subregion)

            for subregion in self.subregions:
                subregion.buildSubRegions()

    cdef void applyForce(self, Node n, double theta, double coefficient=0.0):
        """Apply force from region to a specific node
        
        Args:
            n: Node to apply force to
            theta: Barnes-Hut parameter theta
            coefficient: Repulsion coefficient
        """
        cdef double distance, size_ratio
        
        # If the current node is a leaf (i.e. it has no sub-regions),
        # then there is nothing to do (we don't apply forces from
        # regions onto their own nodes).
        if len(self.nodes) <= 1 or n in self.nodes:
            return
            
        # If the node is a single node, we apply the force directly
        if len(self.nodes) == 1:
            linRepulsion(n, self.nodes[0], coefficient)
            return
            
        # Calculate distance between node and region's mass center
        distance = c_sqrt((n.x - self.massCenterX) ** 2 + (n.y - self.massCenterY) ** 2)
        
        # If distance is zero, we have a problem. Add a small jitter
        if distance < 0.0001:
            distance = 0.0001
            
        # If the region is sufficiently far away or is a leaf, we apply 
        # the force directly using the region's mass center as a proxy
        size_ratio = self.size / distance
        if size_ratio < theta:
            linRepulsion_region(n, self, coefficient)
            return
            
        # Otherwise, we recursively apply the force from the subregions
        for subregion in self.subregions:
            subregion.applyForce(n, theta, coefficient)

    cpdef applyForceOnNodes(self, list nodes, double theta, double coefficient=0.0):
        """Apply forces from this region to a set of nodes
        
        Args:
            nodes: List of nodes to apply forces to
            theta: Barnes-Hut parameter theta
            coefficient: Repulsion coefficient
        """
        cdef Node n
        
        for n in nodes:
            self.applyForce(n, theta, coefficient)


# Adjust speed and apply forces step
cpdef dict adjustSpeedAndApplyForces(list nodes, double speed, double speedEfficiency, double jitterTolerance):
    """Adjust the speed and apply the calculated forces to update node positions
    
    Args:
        nodes: List of nodes
        speed: Current speed
        speedEfficiency: Speed efficiency factor
        jitterTolerance: Jitter tolerance
        
    Returns:
        dict: Dict with updated values for speed and other metrics
    """
    cdef double totalSwinging = 0.0
    cdef double totalEffectiveTraction = 0.0
    cdef Node n
    cdef double swinging, estimatedOptimalJitterTolerance, minJT, maxJT
    cdef double jt, minSpeedEfficiency, targetSpeed, maxRise, factor
    
    # Calculate swinging and effective traction
    for n in nodes:
        swinging = sqrt((n.old_dx - n.dx) * (n.old_dx - n.dx) + (n.old_dy - n.dy) * (n.old_dy - n.dy))
        totalSwinging += n.mass * swinging
        totalEffectiveTraction += n.mass * 0.5 * sqrt((n.old_dx + n.dx) * (n.old_dx + n.dx) + (n.old_dy + n.dy) * (n.old_dy + n.dy))
    
    # Optimize jitter tolerance
    # The 'right' jitter tolerance for this network is totalSwinging / totalEffectiveTraction
    # But given the current jitterTolerance, compute the adaption
    estimatedOptimalJitterTolerance = 0.05 * totalSwinging / totalEffectiveTraction
    
    # Limit to reasonable bounds
    minJT = 0.05
    maxJT = 2.0
    jt = jitterTolerance * max(minJT, min(maxJT, estimatedOptimalJitterTolerance))
    
    minSpeedEfficiency = 0.05
    
    # Protection against erratic behavior
    if totalSwinging / totalEffectiveTraction > 2.0:
        # If swinging is big, it's erratic
        if speedEfficiency > minSpeedEfficiency:
            speedEfficiency *= 0.5
        jt = max(jt, jitterTolerance)
    
    targetSpeed = jt * speedEfficiency * totalEffectiveTraction / totalSwinging
    
    # Speed control
    maxRise = 0.5
    speed = speed + min(targetSpeed - speed, maxRise * speed)
    
    # Apply forces
    for n in nodes:
        swinging = sqrt((n.old_dx - n.dx) * (n.old_dx - n.dx) + (n.old_dy - n.dy) * (n.old_dy - n.dy))
        factor = speed / (1.0 + sqrt(speed * swinging))
        
        n.x = n.x + (n.dx * factor)
        n.y = n.y + (n.dy * factor)
    
    return {
        'speed': speed, 
        'speedEfficiency': speedEfficiency,
        'totalSwinging': totalSwinging,
        'totalEffectiveTraction': totalEffectiveTraction,
        'estimatedOptimalJitterTolerance': estimatedOptimalJitterTolerance
    } 


from itertools import combinations
import matplotlib.pyplot as plt
import matplotlib.patches as mpatches
from matplotlib.patches import Ellipse


#get shared elements for each combination of sets
def get_shared(sets):
    IDs = sets.keys()
    combs = sum([list(map(list, combinations(IDs, i))) for i in range(1, len(IDs) + 1)], [])

    shared = {}
    for comb in combs:
        ID = ' and '.join(comb)
        if len(comb) == 1:
            shared.update({ID: sets[comb[0]]})
        else:
            setlist = [sets[c] for c in comb]
            u = set.intersection(*setlist)
            shared.update({ID: u})
    return shared


#get unique elements for each combination of sets
def get_unique(shared):
    unique = {}
    for shar in shared:
        if shar == list(shared.keys())[-1]:
            s = shared[shar]
            unique.update({shar: s})
            continue
        count = shar.count(' and ')
        if count == 0:
            setlist = [shared[k] for k in shared.keys() if k != shar and " and " not in k]
            s = shared[shar].difference(*setlist)
        else:
            setlist = [shared[k] for k in shared.keys() if k != shar and k.count(' and ') >= count]
            s = shared[shar].difference(*setlist)
        unique.update({shar: s})
    return(unique)


#plot Venn
def venny4py(sets={}, out='./', ce='bgrc',
             asax=False, ext='png', dpi=300, size=3.5,
             bbox_to_anchor=(.5, .99),nc=2,cs=4):
    shared = get_shared(sets)
    unique = get_unique(shared)
    #ce = ce #colors
    lw = size*.12 #line width
    fs = size*2 #font size
    #nc = 2 #legend cols
    #cs = 4 #columnspacing
    
    with open(f'{out}/Intersections_{len(sets)}.txt', 'w') as f:
        for k, v in unique.items():
            f.write(f'{k}: {len(v)}, {sorted(list(v))}\n')
    
    if asax == False:
        plt.rcParams['figure.dpi'] = 80 #dpi in notebook
        plt.rcParams['savefig.dpi'] = dpi #dpi in saved figure
        fig, ax = plt.subplots(1, 1, figsize=(size, size))

    else:
        ax = asax
        
    ax.set_xlim(0, 100)
    ax.set_ylim(0, 100)
    ax.axis('off')
    
    #4 sets
    if len(sets) == 4:
        #draw ellipses
        ew = 45 #width
        eh = 75 #height
        xe = [35, 48, 52, 65] #x coordinats
        ye = [35, 45, 45, 35] #y coordinats
        ae = [225, 225, 315, 315] #angles

        for i, s in enumerate(sets):
            ax.add_artist(Ellipse(xy=(xe[i], ye[i]), width=ew, height=eh, fc=ce[i], 
                                  angle=ae[i], alpha=.3))
            ax.add_artist(Ellipse(xy=(xe[i], ye[i]), width=ew, height=eh, fc='None',
                                  angle=ae[i], ec='black', lw=lw))

        #annotate
        xt = [12, 32, 68, 88, 14, 34, 66, 86, 26, 28, 50, 50, 72, 74, 37, 60, 40, 63, 50] #x
        yt = [67, 79, 79, 67, 41, 70, 70, 41, 59, 26, 11, 60, 26, 59, 51, 17, 17, 51, 35] #y

        for j, s in enumerate(sets):
            ax.text(xt[j], yt[j], len(sets[s]), ha='center', va='center', fontsize=fs, 
                    transform=ax.transData)

        for k in unique:
            j += 1
            ax.text(xt[j], yt[j], len(unique[k]), ha='center', va='center', fontsize=fs, 
                    transform=ax.transData)
            
    #3 sets
    if len(sets) == 3:
        #draw circles
        ew = 60 #width
        eh = 60 #height
        lw = size*.12 #line width
        xe = [37, 63, 50] #x coordinats
        ye = [55, 55, 32] #y coordinats
        nc = 3 #legend columns
        cs = 1 #columns spacing

        for i, s in enumerate(sets):
            ax.add_artist(Ellipse(xy=(xe[i], ye[i]), width=ew, height=eh, fc=ce[i], 
                                  angle=0, alpha=.3))
            ax.add_artist(Ellipse(xy=(xe[i], ye[i]), width=ew, height=eh, fc='None',
                                  angle=0, ec='black', lw=lw))

        #annotate
        xt = [12, 88, 28, 22, 78, 50, 50, 30, 70, 50] #x
        yt = [80, 80,  3, 60, 60, 17, 70, 35, 35, 50] #y
        

        for j, s in enumerate(sets):
            ax.text(xt[j], yt[j], len(sets[s]), ha='center', va='center', fontsize=fs, 
                    transform=ax.transData)

        for k in unique:
            j += 1
            ax.text(xt[j], yt[j], len(unique[k]), ha='center', va='center', fontsize=fs, 
                    transform=ax.transData)
            
    #2 sets
    if len(sets) == 2:
        #draw circles
        ew = 70 #width
        eh = 70 #height
        lw = size*.12 #line width
        xe = [37, 63] #x coordinats
        ye = [45, 45] #y coordinats

        for i, s in enumerate(sets):
            ax.add_artist(Ellipse(xy=(xe[i], ye[i]), width=ew, height=eh, fc=ce[i], 
                                  angle=0, alpha=.3))
            ax.add_artist(Ellipse(xy=(xe[i], ye[i]), width=ew, height=eh, fc='None',
                                  angle=0, ec='black', lw=lw))

        #annotate
        xt = [20, 80, 18, 82, 50] #x
        yt = [80, 80, 45, 45, 45] #y

        for j, s in enumerate(sets):
            ax.text(xt[j], yt[j], len(sets[s]), ha='center', va='center', fontsize=fs, 
                    transform=ax.transData)

        for k in unique:
            j += 1
            ax.text(xt[j], yt[j], len(unique[k]), ha='center', va='center', fontsize=fs, 
                    transform=ax.transData)
                
    #legend
    handles = [mpatches.Patch(color=ce[i], label=l, alpha=.3) for i, l in enumerate(sets)]
    ax.legend(labels=sets, handles=handles, fontsize=fs*1.1, frameon=False, 
              bbox_to_anchor=bbox_to_anchor, bbox_transform=ax.transAxes, loc=9, 
              handlelength=1.5, ncol=nc, columnspacing=cs, handletextpad=.5)
    if asax == False:
        fig.savefig(f'{out}/Venn_{len(sets)}.{ext}', bbox_inches='tight', facecolor='w', )


from .main import CAST_STACK, CAST_MARK, CAST_PROJECT
from .CAST_Stack import reg_params, region_detect, corr_dist, CAST_STACK_rough
from .visualize import kmeans_plot_multiple, plot_mid, dsplot, plot_mid_v2
from .utils import *



import scanpy as sc
import pandas as pd
import anndata
import numpy as np
import time
import numpy as np
from scipy.sparse import issparse, csr_matrix
import igraph
import torch
from tqdm import tqdm
import sys
from ..utils._neighboors import update_rep, eff_n_jobs,neighbors,W_from_rep
from .._settings import add_reference

from ._cosg import cosg
from ..externel.palantir.plot import plot_palantir_results,plot_branch_selection,plot_gene_trends
from ..externel.palantir.utils import run_diffusion_maps,determine_multiscale_space,run_magic_imputation
from ..externel.palantir.core import run_palantir
from ..externel.palantir.presults import select_branch_cells,compute_gene_trends


class TrajInfer(object):
    
    def __init__(self,adata:anndata.AnnData,
                 basis:str='X_umap',use_rep:str='X_pca',n_comps:int=50,
                 n_neighbors:int=15,
                groupby:str='clusters',):
        self.adata=adata
        self.use_rep=use_rep
        self.n_comps=n_comps
        self.basis=basis
        self.groupby=groupby
        self.n_neighbors=n_neighbors
        
        self.origin=None
        self.terminal=None
        
    def set_terminal_cells(self,terminal:list):
        self.terminal=terminal
        
    def set_origin_cells(self,origin:str):
        self.origin=origin
        
    def inference(self,method:str='palantir',**kwargs):
        
        if method=='palantir':

            dm_res = run_diffusion_maps(self.adata,
                                                       pca_key=self.use_rep, 
                                                       n_components=self.n_comps)
            ms_data = determine_multiscale_space(self.adata)
            imputed_X = run_magic_imputation(self.adata)

            sc.tl.rank_genes_groups(self.adata, groupby=self.groupby, 
                        method='t-test',use_rep=self.use_rep,)
            cosg(self.adata, key_added=f'{self.groupby}_cosg', groupby=self.groupby)
            
            ## terminal cells calculation
            if self.terminal is None:
                terminal_states = None  # let Palantir compute terminal cells
            else:
                terminal_index=[]
                for t in self.terminal:
                    gene=sc.get.rank_genes_groups_df(self.adata, group=t, key=f'{self.groupby}_cosg')['names'][0]
                    terminal_index.append(self.adata[self.adata.obs[self.groupby]==t].to_df()[gene].sort_values().index[-1])
                
                terminal_states = pd.Series(
                    self.terminal,
                    index=terminal_index,
                )
            
            ## origin cells calculation
            origin_cell=self.origin
            gene=sc.get.rank_genes_groups_df(self.adata, group=origin_cell, key=f'{self.groupby}_cosg')['names'][0]
            origin_cell_index=self.adata[self.adata.obs[self.groupby]==origin_cell].to_df()[gene].sort_values().index[-1]
            
            start_cell = origin_cell_index
            pr_res = run_palantir(
                self.adata, early_cell=start_cell, terminal_states=terminal_states,
                **kwargs
            )
            
            self.adata.obs['palantir_pseudotime']=pr_res.pseudotime
            add_reference(self.adata,'Palantir','trajectory inference with Palantir')
            return pr_res
        elif method=='diffusion_map':
            sc.pp.neighbors(self.adata, n_neighbors=self.n_neighbors, n_pcs=self.n_comps,
               use_rep=self.use_rep)
            sc.tl.diffmap(self.adata)
            sc.pp.neighbors(self.adata, n_neighbors=self.n_neighbors, use_rep='X_diffmap')
            sc.tl.draw_graph(self.adata)
            self.adata.uns['iroot'] = np.flatnonzero(self.adata.obs[self.groupby]  == self.origin)[0]
            sc.tl.dpt(self.adata)
            sc.pp.neighbors(self.adata, n_neighbors=self.n_neighbors, n_pcs=self.n_comps,
               use_rep=self.use_rep)
            add_reference(self.adata,'diffusion_map','trajectory inference with diffusion map')
        elif method=='slingshot':
            #sc.pp.neighbors(self.adata, n_neighbors=self.n_neighbors, n_pcs=self.n_comps,
            #   use_rep=self.use_rep)
            #sc.tl.umap(self.adata)
            from ._pyslingshot import Slingshot
            slingshot = Slingshot(self.adata, 
                      celltype_key=self.groupby, 
                      obsm_key=self.basis, 
                      start_node=self.origin,
                      end_nodes=self.terminal,
                      debug_level='verbose')
            slingshot.fit(**kwargs)
            pseudotime = slingshot.unified_pseudotime
            self.adata.obs['slingshot_pseudotime']=pseudotime
            self.slingshot=slingshot
            add_reference(self.adata,'slingshot','trajectory inference with slingshot')
        else:
            print('Please input the correct method name, such as `palantir` or `diffusion_map`')
            return
        
    def palantir_plot_pseudotime(self,**kwargs):

        plot_palantir_results(self.adata,**kwargs)
        
    def palantir_cal_branch(self,**kwargs):

        masks = select_branch_cells(self.adata, **kwargs)
        plot_branch_selection(self.adata)

    def palantir_cal_gene_trends(self,layers:str="MAGIC_imputed_data"):

        gene_trends = compute_gene_trends(
            self.adata,
            expression_key=layers,
        )
        return gene_trends
        
    def palantir_plot_gene_trends(self,genes):
        #genes = ['Cdca3','Rasl10a','Mog','Aqp4']

        return plot_gene_trends(self.adata, genes)
    
import networkx as nx
from scipy.sparse import csr_matrix

def construct_graph_(
    W: csr_matrix, directed: bool = False, adjust_weights: bool = True
) -> nx.Graph:
    """
    将稀疏邻接矩阵 W 转换为 networkx 图，并保留权重属性。
    """
    assert issparse(W), "W must be a scipy.sparse matrix"

    # 根据 adjust_weights 选项调整权重
    s, t = W.nonzero()
    w = W.data
    if not directed:
        mask = s < t
        s, t, w = s[mask], t[mask], w[mask]

    if adjust_weights:
        # 标准化并四舍五入到小数点后两位
        w = ((w / np.median(w)) * 100.0 + 0.5).astype(int) / 100.0
        mask = w > 0
        s, t, w = s[mask], t[mask], w[mask]

    # 构建 networkx 图
    G = nx.DiGraph() if directed else nx.Graph()
    G.add_nodes_from(range(W.shape[0]))
    edges = zip(s.tolist(), t.tolist(), w.tolist())
    G.add_weighted_edges_from(edges)
    return G

def construct_graph(
    W, directed: bool = False, adjust_weights: bool = True
) -> "igraph":

    assert issparse(W)

    s, t = W.nonzero()
    w = W.data

    if not directed:
        idx = s < t
        s = s[idx]
        t = t[idx]
        w = w[idx]

    if adjust_weights:
        w = ((w / np.median(w)) * 100.0 + 0.5).astype(
            int
        ) / 100.0  # round to 2 decimal points
        idx = w > 0.0
        if idx.sum() < w.size:
            s = s[idx]
            t = t[idx]
            w = w[idx]

    G = igraph.Graph(directed=directed)
    G.add_vertices(W.shape[0])
    G.add_edges(zip(s, t))
    G.es["weight"] = w

    return G

def calc_force_directed_layout(
    W,
    file_name,
    n_jobs,
    target_change_per_node,
    target_steps,
    is3d,
    memory,
    random_state,
    init=None,
    scaling_ratio=2.0, 
    gravity=1.0, 
    edge_weight_influence=1.0,
    use_gpu=False,  # New parameter to enable GPU acceleration
    optimized_threading=True,  # Enable optimized threading algorithm
    compile_cython=True,  # 是否尝试编译Cython模块以加速
):
    """
    Use our custom ForceAtlas2 implementation to calculate the layout
    """
    G = construct_graph(W)
    try:
        from ..externel.forcedirect2.forceatlas2 import ForceAtlas2
        
        # 如果请求编译并且fa2util模块需要编译
        if compile_cython:
            try:
                import os
                import sys
                import importlib.util
                
                # 检查是否已经编译
                force_dir = os.path.dirname(os.path.abspath(importlib.util.find_spec("omicverse.externel.forcedirect2.forceatlas2").origin))
                compiled_module_exists = any(f.startswith('fa2util.') and (f.endswith('.so') or f.endswith('.pyd')) 
                                           for f in os.listdir(force_dir))
                
                # 如果没有编译过，尝试编译
                if not compiled_module_exists:
                    print("检测到未编译的fa2util模块。尝试编译以获得10-100倍速度提升...")
                    compile_script = os.path.join(force_dir, "compile.py")
                    
                    # 如果compile.py存在，执行它
                    if os.path.exists(compile_script):
                        import subprocess
                        subprocess.run([sys.executable, compile_script], cwd=force_dir)
                    else:
                        print(f"未找到编译脚本: {compile_script}")
                        print("请手动编译fa2util模块以获得更好性能")
                        print("cd " + force_dir)
                        print("python setup.py build_ext --inplace")
            except Exception as e:
                print(f"尝试编译模块时出错: {str(e)}")
                print("将使用未编译的Python版本 (较慢)")
        
        # Initialize ForceAtlas2 with our parameters
        forceatlas2 = ForceAtlas2(
            # Behavior
            outboundAttractionDistribution=True,
            edgeWeightInfluence=edge_weight_influence,
            
            # Performance
            jitterTolerance=1.0,
            barnesHutOptimize=True,
            barnesHutTheta=1.2,
            multiThreaded=True if not use_gpu else False,  # Only use multithreading if not using GPU
            numThreads=n_jobs if n_jobs > 0 else None,  # Use specified number of threads
            useGPU=use_gpu,  # Enable GPU acceleration if requested
            optimizedThreading=optimized_threading,  # Enable optimized threading
            
            # Tuning
            scalingRatio=scaling_ratio,
            strongGravityMode=False,
            gravity=gravity,
            
            # Log
            verbose=True
        )
        
        # Convert igraph to adjacency matrix for forceatlas2
        import numpy as np
        from scipy.sparse import csr_matrix
        
        edges = G.get_edgelist()
        weights = G.es["weight"]
        n_vertices = G.vcount()
        
        rows = [edge[0] for edge in edges]
        cols = [edge[1] for edge in edges]
        
        # Create sparse adjacency matrix
        adj_matrix = csr_matrix((weights, (rows, cols)), shape=(n_vertices, n_vertices))
        
        # Make symmetric if graph is undirected
        if not G.is_directed():
            adj_matrix = adj_matrix + adj_matrix.T
        
        # Run ForceAtlas2
        positions = forceatlas2.forceatlas2(adj_matrix, iterations=target_steps)
        
        # If 3D layout is requested, add a third coordinate with zeros
        if is3d:
            return np.column_stack([positions, np.zeros(len(positions))])
        
        return np.array(positions)
        
    except ImportError:
        # Fallback to external forceatlas2 if our implementation is not available
        import sys
        print("Could not find custom ForceAtlas2 implementation!")
        sys.exit(-1)


def fle(
    data,
    file_name: str = None,
    n_jobs: int = -1,
    rep: str = "diffmap",
    rep_ncomps: int = None,
    K: int = 50,
    full_speed: bool = False,
    target_change_per_node: float = 2.0,
    target_steps: int = 5000,
    is3d: bool = False,
    memory: int = 8,
    random_state: int = 0,
    out_basis: str = "fle",
    fa2_scaling_ratio: float = 2.0,
    fa2_gravity: float = 1.0,
    fa2_edge_weight_influence: float = 1.0,
    use_gpu: bool = False,  # New parameter to enable GPU acceleration
    optimized_threading: bool = True,  # Enable optimized threading algorithm
    compile_cython: bool = True,  # 是否尝试编译Cython模块以加速
) -> None:
    """Construct the Force-directed (FLE) graph.

    This implementation uses our custom ForceAtlas2 implementation, which is a multilthreaded version
    of the original ForceAtlas2 algorithm.

    See [Jacomy14]_ for details on FLE.

    Parameters
    ----------
    data: ``pegasusio.MultimodalData``
        Annotated data matrix with rows for cells and columns for genes.

    file_name: ``str``, optional, default: ``None``
        Temporary file to store the coordinates as the input to forceatlas2. If ``None``, use ``tempfile.mkstemp`` to generate file name.

    n_jobs: ``int``, optional, default: ``-1``
        Number of threads to use. If ``-1``, use all physical CPU cores.

    rep: ``str``, optional, default: ``"diffmap"``
        Representation of data used for the calculation. By default, use Diffusion Map coordinates. If ``None``, use the count matrix ``data.X``.

    rep_ncomps: ``int``, optional (default: None)
        Number of components to be used in `rep`. If rep_ncomps == None, use all components; otherwise, use the minimum of rep_ncomps and rep's dimensions.

    K: ``int``, optional, default: ``50``
        Number of nearest neighbors to be considered during the computation.

    full_speed: ``bool``, optional, default: ``False``
        * If ``True``, use multiple threads in constructing ``hnsw`` index. However, the kNN results are not reproducible.
        * Otherwise, use only one thread to make sure results are reproducible.

    target_change_per_node: ``float``, optional, default: ``2.0``
        Target change per node to stop ForceAtlas2.

    target_steps: ``int``, optional, default: ``5000``
        Maximum number of iterations before stopping the ForceAtlas2 algorithm.

    is3d: ``bool``, optional, default: ``False``
        If ``True``, calculate 3D force-directed layout.

    memory: ``int``, optional, default: ``8``
        Memory size in GB for the Java FA2 component. By default, use 8GB memory.

    random_state: ``int``, optional, default: ``0``
        Random seed set for reproducing results.

    out_basis: ``str``, optional, default: ``"fle"``
        Key name for calculated FLE coordinates to store.
    
    fa2_scaling_ratio: ``float``, optional, default: ``2.0``
        Scaling ratio parameter for ForceAtlas2.
        
    fa2_gravity: ``float``, optional, default: ``1.0``
        Gravity parameter for ForceAtlas2.
        
    fa2_edge_weight_influence: ``float``, optional, default: ``1.0``
        Edge weight influence parameter for ForceAtlas2.
        
    use_gpu: ``bool``, optional, default: ``False``
        Whether to use GPU acceleration for force calculations. Requires PyTorch to be installed.
        
    optimized_threading: ``bool``, optional, default: ``True``
        Whether to use optimized threading algorithm for better parallel performance.
        
    compile_cython: ``bool``, optional, default: ``True``
        Whether to try compiling fa2util module with Cython for 10-100x speedup.

    Returns
    -------
    ``None``

    Update ``data.obsm``:
        * ``data.obsm['X_' + out_basis]``: FLE coordinates of the data.

    Examples
    --------
    >>> pg.fle(data)
    """

    if file_name is None:
        import tempfile

        _, file_name = tempfile.mkstemp()

    rep = update_rep(rep)
    n_jobs = eff_n_jobs(n_jobs)

    if ("W_" + rep) not in data.uns:
        neighbors(
            data,
            K=K,
            rep=rep,
            n_comps=rep_ncomps,
            n_jobs=n_jobs,
            random_state=random_state,
            full_speed=full_speed,
        )

    key = f"X_{out_basis}"
    data.obsm[key] = calc_force_directed_layout(
        W_from_rep(data, rep),
        file_name,
        n_jobs,
        target_change_per_node,
        target_steps,
        is3d,
        memory,
        random_state,
        scaling_ratio=fa2_scaling_ratio,
        gravity=fa2_gravity,
        edge_weight_influence=fa2_edge_weight_influence,
        use_gpu=use_gpu,
        optimized_threading=optimized_threading,
        compile_cython=compile_cython,
    )


def diffmap_fle(
    adata,
    use_rep: str = 'X_pca',
    n_pcs: int = 50,
    n_comps: int = 15,
    n_neighbors: int = 30,
    random_state: int = 0,
    out_basis: str = "fle",
    target_steps: int = 5000,
    use_gpu: bool = False,
    optimized_threading: bool = True,
    compile_cython: bool = True,
    fa2_scaling_ratio: float = 2.0,
    fa2_gravity: float = 1.0,
    fa2_edge_weight_influence: float = 1.0,
    n_jobs: int = -1,
) -> None:
    """Calculate diffusion map and then use it for Force-directed Layout Embedding (FLE).
    
    This function first computes diffusion maps using scanpy and then applies the optimized
    ForceAtlas2 algorithm on the diffusion map coordinates to create a force-directed layout.
    
    Parameters
    ----------
    adata: ``anndata.AnnData``
        Annotated data matrix with rows for cells and columns for genes.
        
    use_rep: ``str``, optional, default: ``'X_pca'``
        The dimensional reduction to use for neighbors calculation before diffmap.
        
    n_pcs: ``int``, optional, default: ``50``
        Number of PCs to use for neighbors calculation.
        
    n_comps: ``int``, optional, default: ``15``
        Number of diffusion components to compute.
        
    n_neighbors: ``int``, optional, default: ``30``
        Number of neighbors for building the neighborhood graph.
        
    random_state: ``int``, optional, default: ``0``
        Random seed for reproducibility.
        
    out_basis: ``str``, optional, default: ``"fle"``
        Key name for calculated FLE coordinates to store.
        
    target_steps: ``int``, optional, default: ``5000``
        Maximum number of iterations for the ForceAtlas2 algorithm.
        
    use_gpu: ``bool``, optional, default: ``False``
        Whether to use GPU acceleration for force calculations.
        
    optimized_threading: ``bool``, optional, default: ``True``
        Whether to use optimized threading algorithm for better parallel performance.
        
    compile_cython: ``bool``, optional, default: ``True``
        Whether to try compiling fa2util module with Cython for speedup.
        
    fa2_scaling_ratio: ``float``, optional, default: ``2.0``
        Scaling ratio parameter for ForceAtlas2.
        
    fa2_gravity: ``float``, optional, default: ``1.0``
        Gravity parameter for ForceAtlas2.
        
    fa2_edge_weight_influence: ``float``, optional, default: ``1.0``
        Edge weight influence parameter for ForceAtlas2.
        
    n_jobs: ``int``, optional, default: ``-1``
        Number of threads to use. If ``-1``, use all physical CPU cores.
        
    Returns
    -------
    ``None``
    
    Update ``adata.obsm``:
        * ``adata.obsm['X_diffmap']``: Diffusion map coordinates
        * ``adata.obsm['X_' + out_basis]``: FLE coordinates
        
    Examples
    --------
    >>> diffmap_fle(adata, use_gpu=True)
    """
    # Step 1: Compute diffusion map using scanpy
    import scanpy as sc
    
    print("Computing neighbors for diffusion map...")
    sc.pp.neighbors(adata, n_neighbors=n_neighbors, n_pcs=n_pcs, 
                   use_rep=use_rep, random_state=random_state)
    
    print(f"Computing diffusion map with {n_comps} components...")
    sc.tl.diffmap(adata, n_comps=n_comps)
    
    # Store the diffusion map connectivity matrix for FLE
    W_diffmap = adata.obsp['connectivities']
    
    # Step 2: Apply ForceAtlas2 on the diffusion map
    print("Calculating force-directed layout from diffusion map...")
    
    import tempfile
    _, file_name = tempfile.mkstemp()
    
    n_jobs = eff_n_jobs(n_jobs)
    
    key = f"X_{out_basis}"
    adata.obsm[key] = calc_force_directed_layout(
        W_diffmap,
        file_name,
        n_jobs,
        2.0,  # target_change_per_node
        target_steps,
        False,  # is3d
        8,  # memory
        random_state,
        scaling_ratio=fa2_scaling_ratio,
        gravity=fa2_gravity,
        edge_weight_influence=fa2_edge_weight_influence,
        use_gpu=use_gpu,
        optimized_threading=optimized_threading,
        compile_cython=compile_cython,
    )
    
    print(f"Force-directed layout calculated and stored in adata.obsm['X_{out_basis}']")
    return None

